from unittest import TestCase
import string
import random
import os
from shutil import copytree, rmtree
import tempfile
import uuid
from aequilibrae.project.network.mode import Mode
from aequilibrae.project import Project
from ...data import no_triggers_project


class TestMode(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(tempfile.gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        copytree(no_triggers_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.curr = self.proj.conn.cursor()

        letters = [random.choice(string.ascii_letters + "_") for x in range(20)]
        self.random_string = "".join(letters)

    def tearDown(self) -> None:
        self.proj.close()

    def test_build(self):
        for val in ["1", "ab", "", None]:
            with self.assertRaises(ValueError):
                m = Mode(val, self.proj)

        for letter in range(10):
            letter = random.choice(string.ascii_letters)
            m = Mode(letter, self.proj)
            del m

    def test_changing_mode_id(self):
        m = Mode("c", self.proj)
        with self.assertRaises(ValueError):
            m.mode_id = "test my description"

    def test_save(self):
        self.curr.execute("select mode_id from 'modes'")

        letter = random.choice([x[0] for x in self.curr.fetchall()])
        m = Mode(letter, self.proj)
        m.mode_name = self.random_string
        m.description = self.random_string[::-1]
        m.save()

        self.curr.execute(f'select description, mode_name from modes where mode_id="{letter}"')

        desc, mname = self.curr.fetchone()
        self.assertEqual(desc, self.random_string[::-1], "Didn't save the mode description correctly")
        self.assertEqual(mname, self.random_string, "Didn't save the mode name correctly")

    def test_empty(self):
        a = Mode("k", self.proj)
        a.mode_name = "just a_test"
        with self.assertRaises(ValueError):
            a.save()

        a = Mode("l", self.proj)
        a.mode_name = "just_a_test_test_with_l"
        with self.assertRaises(ValueError):
            a.save()



import os
import pathlib
import random
import sqlite3
import string
import uuid
from random import choice
from tempfile import gettempdir
from unittest import TestCase

import numpy as np
import pandas as pd

from aequilibrae import TrafficAssignment, TrafficClass, Graph
from aequilibrae.utils.create_example import create_example
from ...data import siouxfalls_project


class TestMCTrafficAssignment(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        proj_path = os.path.join(gettempdir(), "test_mc_traffic_assignment_" + uuid.uuid4().hex)
        self.project = create_example(proj_path)
        self.project.network.build_graphs()
        self.car_graph = self.project.network.graphs["c"]  # type: Graph
        self.truck_graph = self.project.network.graphs["T"]  # type: Graph
        self.moto_graph = self.project.network.graphs["M"]  # type: Graph

        for graph in [self.car_graph, self.truck_graph, self.moto_graph]:
            graph.set_skimming(["free_flow_time"])
            graph.set_graph("free_flow_time")
            graph.set_blocked_centroid_flows(False)

        self.car_matrix = self.project.matrices.get_matrix("demand_mc")
        self.car_matrix.computational_view(["car"])

        self.truck_matrix = self.project.matrices.get_matrix("demand_mc")
        self.truck_matrix.computational_view(["trucks"])

        self.moto_matrix = self.project.matrices.get_matrix("demand_mc")
        self.moto_matrix.computational_view(["motorcycle"])

        self.assignment = TrafficAssignment()
        self.carclass = TrafficClass("car", self.car_graph, self.car_matrix)
        self.carclass.set_pce(1.0)
        self.motoclass = TrafficClass("motorcycle", self.moto_graph, self.moto_matrix)
        self.carclass.set_pce(0.2)
        self.truckclass = TrafficClass("truck", self.truck_graph, self.truck_matrix)
        self.carclass.set_pce(2.5)

        self.algorithms = ["msa", "cfw", "bfw", "frank-wolfe"]

    def tearDown(self) -> None:
        for mat in [self.car_matrix, self.truck_matrix, self.moto_matrix]:
            mat.close()
        self.project.close()

    def test_set_classes(self):
        with self.assertRaises(AttributeError):
            self.assignment.set_classes([1, 2])

        with self.assertRaises(Exception):
            self.assignment.set_classes(self.carclass)

        self.assignment.set_classes([self.carclass, self.truckclass, self.motoclass])

    def test_execute_and_save_results(self):
        self.assignment.set_classes([self.carclass, self.truckclass, self.motoclass])

        for cls in self.assignment.classes:
            cls.graph.set_skimming(["free_flow_time", "distance"])
        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})

        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")

        self.assignment.max_iter = 20
        self.assignment.set_algorithm("bfw")
        self.assignment.execute()

        self.assignment.save_results("save_to_database")
        self.assignment.save_skims(matrix_name="my_skims", which_ones="all")

        with self.assertRaises(ValueError):
            self.assignment.save_results("save_to_database")

        with self.assertRaises(FileExistsError):
            self.assignment.save_skims(matrix_name="my_skims", which_ones="all")



from typing import Union, List, Tuple, Dict
import numpy as np
import pandas as pd

from aequilibrae.paths.graph import Graph
from aequilibrae.matrix import AequilibraeMatrix
from aequilibrae.paths.results import AssignmentResults
import warnings


class TrafficClass:
    """Traffic class for equilibrium traffic assignment

    .. code-block:: python

        >>> from aequilibrae import Project
        >>> from aequilibrae.matrix import AequilibraeMatrix
        >>> from aequilibrae.paths import TrafficClass

        >>> project = Project.from_path("/tmp/test_project")
        >>> project.network.build_graphs()

        >>> graph = project.network.graphs['c'] # we grab the graph for cars
        >>> graph.set_graph('free_flow_time') # let's say we want to minimize time
        >>> graph.set_skimming(['free_flow_time', 'distance']) # And will skim time and distance
        >>> graph.set_blocked_centroid_flows(True)

        >>> proj_matrices = project.matrices

        >>> demand = AequilibraeMatrix()
        >>> demand = proj_matrices.get_matrix("demand_omx")
        >>> demand.computational_view(['matrix'])

        >>> tc = TrafficClass("car", graph, demand)
        >>> tc.set_pce(1.3)
    """

    def __init__(self, name: str, graph: Graph, matrix: AequilibraeMatrix) -> None:
        """
        Instantiates the class

        :Arguments:
            **name** (:obj:`str`): UNIQUE class name.

            **graph** (:obj:`Graph`): Class/mode-specific graph

            **matrix** (:obj:`AequilibraeMatrix`): Class/mode-specific matrix. Supports multiple user classes
        """
        if not np.array_equal(matrix.index, graph.centroids):
            raise ValueError("Matrix and graph do not have compatible sets of centroids.")

        if matrix.matrix_view.dtype != graph.default_types("float"):
            raise TypeError("Matrix's computational view need to be of type np.float64")

        self.graph = graph
        self.logger = graph.logger
        self.matrix = matrix
        self.pce = 1.0
        self.vot = 1.0
        self.mode = graph.mode
        self.class_flow: np.array
        self.results = AssignmentResults()
        self.fixed_cost = np.zeros(graph.graph.shape[0], graph.default_types("float"))
        self.fixed_cost_field = ""
        self.fc_multiplier = 1.0
        self._aon_results = AssignmentResults()
        self._selected_links = {}  # maps human name to link_set
        self.__id__ = name

    def set_pce(self, pce: Union[float, int]) -> None:
        """Sets Passenger Car equivalent

        :Arguments:
            **pce** (:obj:`Union[float, int]`): PCE. Defaults to 1 if not set
        """
        if not isinstance(pce, (float, int)):
            raise ValueError("PCE needs to be either integer or float ")
        self.pce = pce

    def set_fixed_cost(self, field_name: str, multiplier=1):
        """Sets value of time

        :Arguments:
             **field_name** (:obj:`str`): Name of the graph field with fixed costs for this class
             **multiplier** (:obj:`Union[float, int]`): Multiplier for the fixed cost. Defaults to 1 if not set
        """
        if field_name not in self.graph.graph.columns:
            raise ValueError("Field does not exist in the graph")

        self.fc_multiplier = float(multiplier)
        self.fixed_cost_field = field_name
        if np.any(np.isnan(self.graph.graph[field_name].values)):
            self.logger.warning(f"Cost field {field_name} has NaN values. Converted to zero")

        if self.graph.graph[field_name].min() < 0:
            msg = f"Cost field {field_name} has negative values. That is not allowed"
            self.logger.error(msg)
            raise ValueError(msg)

    def set_vot(self, value_of_time: float) -> None:
        """Sets value of time

        :Arguments:
            **value_of_time** (:obj:`Union[float, int]`): Value of time. Defaults to 1 if not set
        """

        self.vot = float(value_of_time)

    def set_select_links(self, links: Dict[str, List[Tuple[int, int]]]):
        """Set the selected links. Checks if the links and directions are valid. Translates link_id and
        direction into unique link id used in compact graph.
        Supply links=None to disable select link analysis.

        :Arguments:
            **links** (:obj:`Union[None, Dict[str, List[Tuple[int, int]]]]`): name of link set and
             Link IDs and directions to be used in select link analysis"""
        self._selected_links = {}
        for name, link_set in links.items():
            if len(name.split(" ")) != 1:
                warnings.warn("Input string name has a space in it. Replacing with _")
                name = str.join("_", name.split(" "))

            link_ids = []
            for link, dir in link_set:
                if dir == 0:
                    query = (self.graph.graph["link_id"] == link) & (
                        (self.graph.graph["direction"] == -1) | (self.graph.graph["direction"] == 1)
                    )
                else:
                    query = (self.graph.graph["link_id"] == link) & (self.graph.graph["direction"] == dir)
                if not query.any():
                    raise ValueError(f"link_id or direction {(link, dir)} is not present within graph.")
                    # Check for duplicate compressed link ids in the current link set
                for comp_id in self.graph.graph[query]["__compressed_id__"].values:
                    if comp_id in link_ids:
                        warnings.warn(
                            "Two input links map to the same compressed link in the network"
                            f", removing superfluous link {link} and direction {dir} with compressed id {comp_id}"
                        )
                    else:
                        link_ids.append(comp_id)
            self._selected_links[name] = np.array(link_ids, dtype=self.graph.default_types("int"))

    def __setattr__(self, key, value):
        if key not in [
            "graph",
            "logger",
            "matrix",
            "pce",
            "mode",
            "class_flow",
            "results",
            "_aon_results",
            "__id__",
            "vot",
            "fixed_cost",
            "fc_multiplier",
            "fixed_cost_field",
            "_selected_links",
        ]:
            raise KeyError("Traffic Class does not have that element")
        self.__dict__[key] = value



import os
import uuid
from shutil import copytree
from tempfile import gettempdir
from unittest import TestCase

import numpy as np

from aequilibrae import Project
from aequilibrae.distribution import Ipf
from aequilibrae.matrix import AequilibraeData
from ...data import siouxfalls_project


class TestIpf(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

    def test_fit(self):
        proj = Project()
        proj.open(self.proj_dir)
        mats = proj.matrices
        mats.update_database()
        seed = mats.get_matrix("SiouxFalls_omx")
        seed.computational_view("matrix")
        # row vector
        args = {"entries": seed.zones, "field_names": ["rows"], "data_types": [np.float64], "memory_mode": True}
        row_vector = AequilibraeData()
        row_vector.create_empty(**args)
        row_vector.rows[:] = np.random.rand(seed.zones)[:] * 1000
        row_vector.index[:] = seed.index[:]
        # column vector
        args["field_names"] = ["columns"]
        column_vector = AequilibraeData()
        column_vector.create_empty(**args)
        column_vector.columns[:] = np.random.rand(seed.zones)[:] * 1000
        column_vector.index[:] = seed.index[:]
        # balance vectors
        column_vector.columns[:] = column_vector.columns[:] * (row_vector.rows.sum() / column_vector.columns.sum())

        # The IPF per se
        args = {
            "matrix": seed,
            "rows": row_vector,
            "row_field": "rows",
            "columns": column_vector,
            "column_field": "columns",
            "nan_as_zero": False,
        }

        with self.assertRaises(TypeError):
            fratar = Ipf(data="test", test="data")
            fratar.fit()

        with self.assertRaises(ValueError):
            fratar = Ipf(**args)
            fratar.parameters = ["test"]
            fratar.fit()

        fratar = Ipf(**args)
        fratar.fit()

        result = fratar.output

        self.assertAlmostEqual(
            np.nansum(result.matrix_view), np.nansum(row_vector.data["rows"]), 4, "Ipf did not converge"
        )
        self.assertGreater(fratar.parameters["convergence level"], fratar.gap, "Ipf did not converge")

        mr = fratar.save_to_project("my_matrix_ipf", "my_matrix_ipf.aem")

        self.assertTrue(
            os.path.isfile(os.path.join(mats.fldr, "my_matrix_ipf.aem")), "Did not save file to the appropriate place"
        )

        self.assertEqual(mr.procedure_id, fratar.procedure_id, "procedure ID saved wrong")
        proj.close()



import pytest
import os


@pytest.fixture
def pat(create_path, create_gtfs_project):
    gtfs_fldr = os.path.join(create_path, "gtfs_coquimbo.zip")

    transit = create_gtfs_project.new_gtfs_builder(
        agency="LISERCO, LISANCO, LINCOSUR", file_path=gtfs_fldr, description=""
    )
    transit.load_date("2016-04-13")

    patterns = transit.select_patterns
    yield [x for x in patterns.values()][0]


def test_save_to_database(pat, transit_conn):
    pat.save_to_database(transit_conn)

    routes = transit_conn.execute("SELECT COUNT(*) FROM routes;").fetchone()[0]
    assert routes == 1


def test_best_shape(pat):
    shp = pat.best_shape()
    assert shp == pat._stop_based_shape, "Returned the wrong shape"


def test_get_error(pat):
    assert pat.get_error() is None, "Resulted a map-matching error when should have returned none"


def test_map_match(pat, transit_conn):
    pat.map_match()
    pat.save_to_database(transit_conn)

    pattern_map = transit_conn.execute("SELECT COUNT(*) FROM pattern_mapping;").fetchone()[0]
    assert pattern_map > 0



import os
import sqlite3
import tempfile
import uuid

import numpy as np
from numpy.lib.format import open_memmap

MEMORY = 1
DISK = 0
data_export_types = ["aed", "csv", "sqlite"]


class AequilibraeData(object):
    """AequilibraE dataset"""

    def __init__(self):
        self.data = None
        self.file_path = None
        self.entries = None
        self.fields = None
        self.num_fields = None
        self.data_types = None
        self.aeq_index_type = None
        self.memory_mode = None

    @classmethod
    def empty(cls, *args, **kwargs):
        instance = cls()
        instance.create_empty(*args, **kwargs)
        return instance

    def create_empty(
        self, file_path=None, entries=1, field_names=None, data_types=None, memory_mode=False, fill=None, index=None
    ):
        """
        Creates a new empty dataset

        :Arguments:
            **file_path** (:obj:`str`, Optional): Full path for the output data file. If *memory_mode* is 'false' and
            path is missing, then the file is created in the temp folder

            **entries** (:obj:`int`, Optional): Number of records in the dataset. Default is 1

            **field_names** (:obj:`list`, Optional): List of field names for this dataset. If no list is provided, the
            field 'data' will be created

            **data_types** (:obj:`np.dtype`, Optional): List of data types for the dataset. Types need to be NumPy data
            types (e.g. np.int16, np.float64). If no list of types are provided, type will be *np.float64*

            **memory_mode** (:obj:`bool`, Optional): If true, dataset will be kept in memory. If false, the dataset will
            be a memory-mapped numpy array

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeData, AequilibraeMatrix

            >>> mat = AequilibraeMatrix()
            >>> mat.load('/tmp/test_project/matrices/demand.omx')
            >>> mat.computational_view()

            >>> vectors = "/tmp/test_project/vectors.aed"

            >>> args = {
            ...      "file_path": vectors,
            ...      "entries": mat.zones,
            ...      "field_names": ["origins", "destinations"],
            ...      "data_types": [np.float64, np.float64]
            ... }

            >>> dataset = AequilibraeData()
            >>> dataset.create_empty(**args)

        """

        if file_path is not None or memory_mode:
            if field_names is None:
                field_names = ["data"]

            if data_types is None:
                data_types = [np.float64] * len(field_names)

            self.file_path = file_path
            self.entries = entries
            self.fields = field_names
            self.data_types = data_types
            self.aeq_index_type = np.uint64

            if memory_mode:
                self.memory_mode = MEMORY
            else:
                self.memory_mode = DISK
                if self.file_path is None:
                    self.file_path = self.random_name()

            # Consistency checks
            if not isinstance(self.fields, list):
                raise ValueError('Titles for fields, "field_names", needs to be a list')

            if not isinstance(self.data_types, list):
                raise ValueError('Data types, "data_types", needs to be a list')
            # The check below is not working properly with the QGIS importer
            # else:
            #     for dt in self.data_types:
            #         if not isinstance(dt, type):
            #             raise ValueError('Data types need to be Python or Numpy data types')

            for field in self.fields:
                if field in object.__dict__:
                    raise Exception(field + " is a reserved name. You cannot use it as a field name")

            self.num_fields = len(self.fields)

            dtype = [("index", self.aeq_index_type)] + [(f, dt) for f, dt in zip(self.fields, self.data_types)]

            # the file
            if self.memory_mode:
                self.data = np.recarray((self.entries,), dtype=dtype)
            else:
                self.data = open_memmap(self.file_path, mode="w+", dtype=dtype, shape=(self.entries,))

            if fill is not None:
                [self.data[f].fill(fill) for f in self.fields]
            if index is not None:
                self.index[:] = index[:]

    def __getattr__(self, field_name):
        if field_name in object.__dict__:
            return self.__dict__[field_name]

        if self.data is not None:
            if field_name in self.fields:
                return self.data[field_name]

            if field_name == "index":
                return self.data["index"]

            raise AttributeError("No such method or data field! --> " + str(field_name))
        else:
            raise AttributeError("Data container is empty")

    def load(self, file_path):
        """
        Loads dataset from file

        :Arguments:
            **file_path** (:obj:`str`): Full file path to the AequilibraeData to be loaded

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeData

            >>> dataset = AequilibraeData()
            >>> dataset.load("/tmp/test_project/vectors.aed")
        """
        f = open(file_path)
        self.file_path = os.path.realpath(f.name)
        f.close()

        # Map in memory and load data names plus dimensions
        self.data = open_memmap(self.file_path, mode="r+")

        self.entries = self.data.shape[0]
        self.fields = [x for x in self.data.dtype.fields if x != "index"]
        self.num_fields = len(self.fields)
        self.data_types = [self.data[x].dtype.type for x in self.fields]

    def export(self, file_name, table_name="aequilibrae_table"):
        """
        Exports the dataset to another format. Supports CSV and SQLite

        :Arguments:
            **file_name** (:obj:`str`): File name with PATH and extension (csv, or sqlite3, sqlite or db)

            **table_name** (:obj:`str`): It only applies if you are saving to an SQLite table. Otherwise ignored

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeData

            >>> dataset = AequilibraeData()
            >>> dataset.load("/tmp/test_project/vectors.aed")
            >>> dataset.export("/tmp/test_project/vectors.csv")
        """

        file_type = os.path.splitext(file_name)[1]
        headers = ["index"]
        headers.extend(self.fields)

        if file_type.lower() == ".aed":
            dtype = [("index", self.aeq_index_type)]
            dtype.extend([(self.fields[i], self.data_types[i]) for i in range(self.num_fields)])
            data = open_memmap(file_name, mode="w+", dtype=dtype, shape=(self.entries,))
            for field in data.dtype.names:
                data[field] = self.data[field]
            data.flush()
            del data

        elif file_type.lower() == ".csv":
            fmt = "%d"
            for dt in self.data_types:
                if np.issubdtype(dt, np.floating):
                    fmt += ",%f"
                elif np.issubdtype(dt, np.integer):
                    fmt += ",%d"
            data = np.array(self.data, copy=True)
            for nm in self.data.dtype.names:
                np.nan_to_num(data[nm], copy=False)

            np.savetxt(file_name, data[np.newaxis, :][0], delimiter=",", fmt=fmt, header=",".join(headers), comments="")

        elif file_type.lower() in [".sqlite", ".sqlite3", ".db"]:
            # Connecting to the database file
            conn = sqlite3.connect(file_name)
            c = conn.cursor()
            # Creating the table, but before deletes if the table exists
            c.execute("""DROP TABLE IF EXISTS """ + table_name)
            fi = ""
            qm = "?"
            for f in headers[1:]:
                fi += ", " + f + " REAL"
                qm += ", ?"

            c.execute("""CREATE TABLE """ + table_name + """ (link_id INTEGER PRIMARY KEY""" + fi + ")" "")
            c.execute("BEGIN TRANSACTION")
            c.executemany("INSERT INTO " + table_name + " VALUES (" + qm + ")", self.data)
            c.execute("END TRANSACTION")
            conn.commit()
            conn.close()

    @staticmethod
    def random_name():
        """
        Returns a random name for a dataset with root in the temp directory of the user

        .. code-block:: python

            >>> from aequilibrae.matrix import AequilibraeData

            >>> name = AequilibraeData().random_name() # doctest: +ELLIPSIS

            # This is an example of output
            # '/tmp/Aequilibrae_data_5werr5f36-b123-asdf-4587-adfglkjhqwe.aed'
        """
        return os.path.join(tempfile.gettempdir(), f"Aequilibrae_data_{uuid.uuid4()}.aed")



from unittest import TestCase
from tempfile import gettempdir
import os
import uuid
from shutil import copytree
from aequilibrae.project import Project
from warnings import warn
from random import random
from ...data import siouxfalls_project


class TestNetwork(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.proj_path = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_path)
        self.siouxfalls = Project()
        self.siouxfalls.open(self.proj_path)
        self.proj_path2 = os.path.join(gettempdir(), uuid.uuid4().hex)

    def tearDown(self) -> None:
        self.siouxfalls.close()

    def test_create_from_osm(self):
        thresh = 0.05
        if os.environ.get("GITHUB_WORKFLOW", "ERROR") == "Code coverage":
            thresh = 1.01

        if random() < thresh:
            self.siouxfalls.close()
            self.project = Project()
            self.project.new(self.proj_path2)
            # self.network.create_from_osm(west=153.1136245, south=-27.5095487, east=153.115, north=-27.5085, modes=["car"])
            self.project.network.create_from_osm(west=-112.185, south=36.59, east=-112.179, north=36.60)
            curr = self.project.conn.cursor()

            curr.execute("""select count(*) from links""")
            lks = curr.fetchone()[0]

            curr.execute("""select count(distinct osm_id) from links""")
            osmids = curr.fetchone()[0]

            if osmids == 0:
                warn("COULD NOT RETRIEVE DATA FROM OSM")
                return

            if osmids >= lks:
                self.fail("OSM links not broken down properly")

            curr.execute("""select count(*) from nodes""")
            nds = curr.fetchone()[0]

            if lks > nds:
                self.fail("We imported more links than nodes. Something wrong here")
            self.project.close()
            self.siouxfalls.open(self.proj_path)
        else:
            print("Skipped check to not load OSM servers")

    def test_count_centroids(self):
        items = self.siouxfalls.network.count_centroids()
        self.assertEqual(24, items, "Wrong number of centroids found")

        nodes = self.siouxfalls.network.nodes
        node = nodes.get(1)
        node.is_centroid = 0
        node.save()

        items = self.siouxfalls.network.count_centroids()
        self.assertEqual(23, items, "Wrong number of centroids found")

    def test_count_links(self):
        items = self.siouxfalls.network.count_links()
        self.assertEqual(76, items, "Wrong number of links found")

    def test_count_nodes(self):
        items = self.siouxfalls.network.count_nodes()
        self.assertEqual(24, items, "Wrong number of nodes found")



from shapely.geometry import LineString
from aequilibrae.project import Project
from aequilibrae.transit import Transit
import pytest

from aequilibrae.project.database_connection import database_connection


class TestTransitTriggers:
    def test_link_insert(self, project: Project):
        Transit(project)

        conn = database_connection(db_type="transit")

        data = [10001001000, 3, 20000001, 5, 6, 0, LineString([[-23.59, -46.64], [-23.43, -46.50]]).wkb]
        conn.execute(
            """INSERT INTO route_links (pattern_id, seq, transit_link, from_stop, to_stop, distance, geometry)
                        VALUES(?, ?, ?, ?, ?, ?, GeomFromWKB(?, 4326));""",
            data,
        )
        conn.commit()

        distance = conn.execute("SELECT distance FROM route_links WHERE seq=3;").fetchone()[0]

        assert distance != 0

    def test_geometry_update(self, project: Project):
        Transit(project)

        conn = database_connection(db_type="transit")

        data = [10001001000, 3, 20000001, 5, 6, 0, LineString([[-23.59, -46.64], [-23.43, -46.50]]).wkb]
        conn.execute(
            """INSERT INTO route_links (pattern_id, seq, transit_link, from_stop, to_stop, distance, geometry)
                        VALUES(?, ?, ?, ?, ?, ?, GeomFromWKB(?, 4326));""",
            data,
        )
        conn.commit()

        conn.execute(
            "UPDATE route_links SET geometry=GeomFromWKB(?, 4326) WHERE seq=3;",
            [LineString([[-23.59, -46.64], [-23.01, -47.14]]).wkb],
        )
        conn.commit()

        distance = conn.execute("SELECT distance FROM route_links WHERE seq=3;").fetchone()[0]

        assert round(distance, 2) != 19815.63



from aequilibrae.context import get_active_project
import pytest


class TestMultipleProjects:
    @pytest.fixture(scope="class")
    def project(self, create_empty_project_session):
        return create_empty_project_session()

    def test_current_project_is_active_project(self, project):
        assert project is get_active_project()

    def test_switch_project(self, create_empty_project):
        proj2 = create_empty_project()
        assert proj2 is get_active_project()

    def test_reactivate_project(self, project, create_empty_project):
        create_empty_project()
        project.activate()
        assert project is get_active_project()

    def test_raises_when_inactive(self, project):
        project.deactivate()
        with pytest.raises(FileNotFoundError):
            get_active_project()

    def test_close_project_deactivates(self, project):
        project.close()
        assert get_active_project(must_exist=False) is None



from unittest import TestCase
from tempfile import gettempdir
from os.path import join
from uuid import uuid4
from aequilibrae.utils.create_example import create_example
from aequilibrae.utils.create_delaunay_network import DelaunayAnalysis


class TestDelaunayAnalysis(TestCase):
    def setUp(self) -> None:
        self.proj = create_example(join(gettempdir(), uuid4().hex))

    def tearDown(self) -> None:
        self.proj.close()

    def test_create_delaunay_network(self):
        da = DelaunayAnalysis(self.proj)
        with self.assertRaises(ValueError):
            da.create_network("nodes")

        da.create_network()
        self.assertEqual(59, self.proj.conn.execute("select count(*) from delaunay_network").fetchone()[0])

        da.create_network("network", True)
        self.assertEqual(62, self.proj.conn.execute("select count(*) from delaunay_network").fetchone()[0])

        with self.assertRaises(ValueError):
            da.create_network()

    def test_assign_matrix(self):
        demand = self.proj.matrices.get_matrix("demand_omx")
        demand.computational_view(["matrix"])
        da = DelaunayAnalysis(self.proj)
        da.create_network()
        da.assign_matrix(demand, "delaunay_test")



import os
import uuid
from copy import copy, deepcopy
from random import randint, random
from shutil import copytree, rmtree
from tempfile import gettempdir
from unittest import TestCase, TestLoader as _TestLoader

import shapely.wkb
from shapely.geometry import Point

from aequilibrae.project import Project
from ...data import siouxfalls_project

_TestLoader.sortTestMethodsUsing = None


class TestNodes(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        self.proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.proj_dir)

        self.project = Project()
        self.project.open(self.proj_dir)
        self.network = self.project.network
        self.curr = self.project.conn.cursor()

    def tearDown(self) -> None:
        self.project.close()
        del self.curr
        try:
            rmtree(self.proj_dir)
        except Exception as e:
            print(f"Failed to remove at {e.args}")

    def test_get(self):
        nodes = self.network.nodes
        nd = randint(1, 24)
        node = nodes.get(nd)

        self.assertEqual(node.node_id, nd, "get node returned wrong object")

        # Make sure that if we renumber itg we would not get it again
        node.renumber(200)
        with self.assertRaises(ValueError):
            node = nodes.get(nd)

    def test_save(self):
        nodes = self.network.nodes
        chosen = [randint(1, 24) for _ in range(5)]
        while len(chosen) != len(set(chosen)):
            chosen = [randint(1, 24) for _ in range(5)]
        coords = []
        for nd in chosen:
            node = nodes.get(nd)
            node.is_centroid = 0
            x = node.geometry.x + random()
            y = node.geometry.y + random()
            coords.append([x, y])
            node.geometry = Point([x, y])

        nodes.save()
        for nd, coords in zip(chosen, coords):
            x, y = coords
            self.curr.execute("Select is_centroid, asBinary(geometry) from nodes where node_id=?;", [nd])
            flag, wkb = self.curr.fetchone()
            self.assertEqual(flag, 0, "Saving of is_centroid failed")

            geo = shapely.wkb.loads(wkb)
            self.assertEqual(geo.x, x, "Geometry X saved wrong")
            self.assertEqual(geo.y, y, "Geometry Y saved wrong")

    def test_fields(self):
        nodes = self.network.nodes
        f_editor = nodes.fields

        fields = sorted(f_editor.all_fields())
        self.curr.execute("pragma table_info(nodes)")
        dt = self.curr.fetchall()

        actual_fields = set([x[1] for x in dt if x[1] != "ogc_fid"])
        actual_fields = sorted(list(actual_fields))

        self.assertEqual(fields, actual_fields, "Table editor is weird for table nodes")

    def test_copy(self):
        nodes = self.network.nodes
        with self.assertRaises(Exception):
            _ = copy(nodes)
        with self.assertRaises(Exception):
            _ = deepcopy(nodes)

    def test_new_centroid(self):
        nodes = self.network.nodes

        with self.assertRaises(Exception):
            node = nodes.new_centroid(1)

        tot_prev_centr = self.network.count_centroids()
        tot_prev_nodes = self.network.count_nodes()
        node = nodes.new_centroid(100)
        self.assertEqual(1, node.is_centroid, "Creating new centroid returned wrong is_centroid value")
        node.geometry = Point(1, 1)
        node.save()

        self.assertEqual(tot_prev_centr + 1, self.network.count_centroids(), "Failed to add centroids")
        self.assertEqual(tot_prev_nodes + 1, self.network.count_nodes(), "Failed to add centroids")



from tempfile import gettempdir
import urllib.request
import platform
import zipfile
from os.path import join
from os import walk

pth = "https://github.com/AequilibraE/aequilibrae/releases/download/V0.6.0.post1/mod_spatialite-NG-win-amd64.zip"

outfolder = gettempdir()

dest_path = join(outfolder, "mod_spatialite-NG-win-amd64.zip")
urllib.request.urlretrieve(pth, dest_path)

fldr = join(outfolder, "temp_data")
zipfile.ZipFile(dest_path).extractall(fldr)

if "WINDOWS" in platform.platform().upper():
    # We now set sqlite. Only needed in thge windows server in Github
    plats = {
        "x86": "https://sqlite.org/2020/sqlite-dll-win32-x86-3320100.zip",
        "x64": "https://sqlite.org/2020/sqlite-dll-win64-x64-3320100.zip",
    }

    outfolder = "C:/"
    zip_path64 = join(outfolder, "sqlite-dll-win64-x64-3320100.zip")
    urllib.request.urlretrieve(plats["x64"], zip_path64)

    zip_path86 = join(outfolder, "sqlite-dll-win32-x86-3320100.zip")
    urllib.request.urlretrieve(plats["x86"], zip_path86)

    root = "C:/hostedtoolcache/windows/Python/"
    file = "sqlite3.dll"
    for d, subD, f in walk(root):
        if file in f:
            if "x64" in d:
                zipfile.ZipFile(zip_path64).extractall(d)
            else:
                zipfile.ZipFile(zip_path86).extractall(d)
            print(f"Replaces {d}")



from unittest import TestCase
from tempfile import gettempdir
from uuid import uuid4
from os.path import join
from aequilibrae.paths import TrafficClass
from aequilibrae.paths import Graph
from aequilibrae.paths.results import AssignmentResults
from aequilibrae.utils.create_example import create_example


class TestTrafficClass(TestCase):
    def setUp(self) -> None:
        self.project = create_example(join(gettempdir(), "test_set_pce_" + uuid4().hex))
        self.project.network.build_graphs()
        car_graph = self.project.network.graphs["c"]  # type: Graph
        car_graph.set_graph("distance")
        car_graph.set_blocked_centroid_flows(False)
        matrix = self.project.matrices.get_matrix("demand_omx")
        matrix.computational_view()
        self.tc = TrafficClass(name="car", graph=car_graph, matrix=matrix)

    def tearDown(self) -> None:
        self.project.close()

    def test_result_type(self):
        self.assertIsInstance(self.tc.results, AssignmentResults, "Results have the wrong type")
        self.assertIsInstance(self.tc._aon_results, AssignmentResults, "Results have the wrong type")

    def test_set_pce(self):
        with self.assertRaises(ValueError):
            self.tc.set_pce("not a number")
        self.tc.set_pce(1)
        self.tc.set_pce(3.9)

    def test_set_vot(self):
        self.assertEqual(self.tc.vot, 1.0)
        self.tc.set_vot(4.5)
        self.assertEqual(self.tc.vot, 4.5)

    def test_set_fixed_cost(self):
        self.assertEqual(self.tc.fc_multiplier, 1.0)

        with self.assertRaises(ValueError):
            self.tc.set_fixed_cost("Field_Does_Not_Exist", 2.5)
        self.assertEqual(self.tc.fc_multiplier, 1.0)

        self.tc.set_fixed_cost("distance", 3.0)
        self.assertEqual(self.tc.fc_multiplier, 3.0)
        self.assertEqual(self.tc.fixed_cost_field, "distance")



import os
import pathlib
import uuid
import zipfile
from os.path import join, dirname
from shutil import copytree
from tempfile import gettempdir
from unittest import TestCase
import numpy as np
import pandas as pd

from aequilibrae import TrafficAssignment, TrafficClass, Graph, Project, PathResults
from ...data import siouxfalls_project


class TestSelectLink(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

        proj_path = os.path.join(gettempdir(), "test_traffic_assignment_path_files" + uuid.uuid4().hex)
        os.mkdir(proj_path)
        zipfile.ZipFile(join(dirname(siouxfalls_project), "sioux_falls_single_class.zip")).extractall(proj_path)
        self.project = Project()
        self.project.open(proj_path)
        self.project.network.build_graphs()
        self.car_graph = self.project.network.graphs["c"]  # type: Graph
        self.car_graph.set_graph("free_flow_time")
        self.car_graph.set_blocked_centroid_flows(False)
        self.matrix = self.project.matrices.get_matrix("demand_omx")
        self.matrix.computational_view()

        self.assignment = TrafficAssignment()
        self.assignclass = TrafficClass("car", self.car_graph, self.matrix)
        self.assignment.set_classes([self.assignclass])
        self.assignment.set_vdf("BPR")
        self.assignment.set_vdf_parameters({"alpha": 0.15, "beta": 4.0})
        self.assignment.set_vdf_parameters({"alpha": "b", "beta": "power"})
        self.assignment.set_capacity_field("capacity")
        self.assignment.set_time_field("free_flow_time")
        self.assignment.max_iter = 1
        self.assignment.set_algorithm("msa")

    def tearDown(self) -> None:
        self.matrix.close()
        self.project.close()

    def test_multiple_link_sets(self):
        """
        Tests whether the Select Link feature works as wanted.
        Uses two examples: 2 links in one select link, and a single Selected Link
        Checks both the OD Matrix and Link Loading
        """
        self.assignclass.set_select_links({"9 or 6": [(9, 1), (6, 1)], "just 3": [(3, 1)], "5 for fun": [(5, 1)]})
        self.assignment.execute()
        for key in self.assignclass._selected_links.keys():
            od_mask, link_loading = create_od_mask(
                self.assignclass.matrix.matrix_view, self.assignclass.graph, self.assignclass._selected_links[key]
            )
            np.testing.assert_allclose(
                self.assignclass.results.select_link_od.matrix[key][:, :, 0],
                od_mask,
                err_msg="OD SL matrix for: " + str(key) + " does not match",
            )
            np.testing.assert_allclose(
                self.assignclass.results.select_link_loading[key],
                link_loading,
                err_msg="Link loading SL matrix for: " + str(key) + " does not match",
            )

    def test_equals_demand_one_origin(self):
        """
        Test to ensure the Select Link functionality behaves as required.
        Tests to make sure the OD matrix works when all links surrounding one origin are selected
        Confirms the Link Loading is done correctly in this case
        """
        self.assignclass.set_select_links({"1, 4, 3, and 2": [(1, 1), (4, 1), (3, 1), (2, 1)]})

        self.assignment.execute()

        for key in self.assignclass._selected_links.keys():
            od_mask, link_loading = create_od_mask(
                self.assignclass.matrix.matrix_view, self.assignclass.graph, self.assignclass._selected_links[key]
            )
            np.testing.assert_allclose(
                self.assignclass.results.select_link_od.matrix[key][:, :, 0],
                od_mask,
                err_msg="OD SL matrix for: " + str(key) + " does not match",
            )
            np.testing.assert_allclose(
                self.assignclass.results.select_link_loading[key],
                link_loading,
                err_msg="Link loading SL matrix for: " + str(key) + " does not match",
            )

    def test_single_demand(self):
        """
        Tests the functionality of Select Link when given a custom demand matrix, where only 1 OD pair has demand on it
        Confirms the OD matrix behaves, and the Link Loading is just on the path of this OD pair
        """
        custom_demand = np.zeros((24, 24, 1))
        custom_demand[0, 23, 0] = 1000
        self.matrix.matrix_view = custom_demand
        self.assignclass.matrix = self.matrix

        self.assignclass.set_select_links({"39, 66, or 73": [(39, 1), (66, 1), (73, 1)]})

        self.assignment.execute()
        for key in self.assignclass._selected_links.keys():
            od_mask, link_loading = create_od_mask(
                self.assignclass.matrix.matrix_view, self.assignclass.graph, self.assignclass._selected_links[key]
            )
            np.testing.assert_allclose(
                self.assignclass.results.select_link_od.matrix[key][:, :, 0],
                od_mask,
                err_msg="OD SL matrix for: " + str(key) + " does not match",
            )
            np.testing.assert_allclose(
                self.assignclass.results.select_link_loading[key],
                link_loading,
                err_msg="Link loading SL matrix for: " + str(key) + " does not match",
            )

    def test_select_link_network_loading(self):
        """
        Test to ensure the SL_network_loading method correctly does the network loading
        """
        self.assignment.execute()
        non_sl_loads = self.assignclass.results.get_load_results()
        self.setUp()
        self.assignclass.set_select_links({"39, 66, or 73": [(39, 1), (66, 1), (73, 1)]})
        self.assignment.execute()
        sl_loads = self.assignclass.results.get_load_results()
        np.testing.assert_allclose(non_sl_loads.matrix_tot, sl_loads.matrix_tot)

    def test_duplicate_links(self):
        """
        Tests to make sure the user api correctly filters out duplicate links in the compressed graph
        """
        self.assignment = TrafficAssignment()
        self.assignclass = TrafficClass("car", self.car_graph, self.matrix)
        self.assignclass.set_select_links({"test": [(1, 1), (1, 1)]})
        self.assertEquals(len(self.assignclass._selected_links["test"]), 1, "Did not correctly remove duplicate link")

    def test_link_out_of_bounds(self):
        """
        Test to confirm the user api correctly identifies when an input node is invalid for the current graph
        """
        self.assignment = TrafficAssignment()
        self.assignclass = TrafficClass("car", self.car_graph, self.matrix)
        self.assertRaises(ValueError, self.assignclass.set_select_links, {"test": [(78, 1), (1, 1)]})


def create_od_mask(demand: np.array, graph: Graph, sl):
    res = PathResults()
    # This uses the UNCOMPRESSED graph, since we don't know which nodes the user may ask for
    graph.set_graph("free_flow_time")
    res.prepare(graph)

    def g(o, d):
        res.compute_path(o, d)
        return list(res.path_nodes) if (res.path_nodes is not None and o != d) else []

    a = [[g(o, d) for d in range(1, 25)] for o in range(1, 25)]
    sl_links = []
    for i in range(len(sl)):
        node_pair = graph.graph.iloc[sl[i]]["a_node"] + 1, graph.graph.iloc[sl[i]]["b_node"] + 1
        sl_links.append(node_pair)
    mask = dict()
    for origin, val in enumerate(a):
        for dest, path in enumerate(val):
            for k in range(1, len(path)):
                if origin == dest:
                    pass
                elif (path[k - 1], path[k]) in sl_links:
                    mask[(origin, dest)] = True
    sl_od = np.zeros((24, 24))
    for origin in range(24):
        for dest in range(24):
            if mask.get((origin, dest)):
                sl_od[origin, dest] = demand[origin, dest]

    # make link loading
    loading = np.zeros((76, 1))
    for orig, dest in mask.keys():
        path = a[orig][dest]
        for i in range(len(path) - 1):
            link = (
                graph.graph[(graph.graph["a_node"] == path[i] - 1) & (graph.graph["b_node"] == path[i + 1] - 1)][
                    "link_id"
                ].values[0]
                - 1
            )
            loading[link] += demand[orig, dest]
    return sl_od, loading



from unittest import TestCase
import os
from shutil import copytree, rmtree
import uuid
import random
import string
import tempfile
from sqlite3 import IntegrityError
from aequilibrae import Project
from aequilibrae.project.network.mode import Mode
from ...data import siouxfalls_project


class TestModes(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(tempfile.gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.curr = self.proj.conn.cursor()

    def tearDown(self) -> None:
        self.proj.close()

    def test_add(self):
        new_mode = Mode("F", self.proj)
        name = [random.choice(string.ascii_letters + "_") for x in range(random.randint(1, 20))]
        name = "".join(name)
        new_mode.mode_name = name
        self.proj.network.modes.add(new_mode)

        self.curr.execute('select mode_name from modes where mode_id="F"')
        self.assertEqual(self.curr.fetchone()[0], name, "Could not save the mode properly to the database")

    def test_drop(self):
        self.proj.network.modes.delete("b")

        with self.assertRaises(IntegrityError):
            self.proj.network.modes.delete("c")

    def test_get(self):
        c = self.proj.network.modes.get("c")
        self.assertEqual("All motorized vehicles", c.description)
        del c

        with self.assertRaises(ValueError):
            _ = self.proj.network.modes.get("f")

    def test_new(self):
        modes = self.proj.network.modes
        self.assertIsInstance(modes.new("h"), Mode, "Returned wrong type")

        m = list(modes.all_modes().keys())[0]
        with self.assertRaises(ValueError):
            modes.new(m)

    def test_fields(self):
        fields = self.proj.network.modes.fields
        fields.all_fields()
        self.assertEqual(fields._table, "modes", "Returned wrong table handler")



from copy import deepcopy

import pandas as pd

from aequilibrae.project.basic_table import BasicTable
from aequilibrae.project.data_loader import DataLoader
from aequilibrae.project.network.node import Node
from aequilibrae.project.table_loader import TableLoader


class Nodes(BasicTable):
    """
    Access to the API resources to manipulate the links table in the network

    .. code-block:: python

        >>> from aequilibrae import Project

        >>> proj = Project.from_path("/tmp/test_project")

        >>> all_nodes = proj.network.nodes

        # We can just get one link in specific
        >>> node = all_nodes.get(21)

        # We can save changes for all nodes we have edited so far
        >>> all_nodes.save()
    """

    #: Query sql for retrieving nodes
    sql = ""

    def __init__(self, net):
        super().__init__(net.project)
        self.__table_type__ = "nodes"
        self.__items = {}
        self.__fields = []

        if self.sql == "":
            self.refresh_fields()

    def get(self, node_id: int) -> Node:
        """Get a node from the network by its **node_id**

        It raises an error if node_id does not exist

        :Arguments:
            **node_id** (:obj:`int`): Id of a node to retrieve

        :Returns:
            **node** (:obj:`Node`): Node object for requested node_id
        """

        if node_id in self.__items:
            node = self.__items[node_id]

            # If this element has not been renumbered, we return it. Otherwise we
            # store the object under its new number and carry on
            if node.node_id == node_id:
                return node
            else:
                self.__items[node.node_id] = self.__items.pop(node_id)

        self._curr.execute(f"{self.sql} where node_id=?", [node_id])
        data = self._curr.fetchone()
        if data:
            data = {key: val for key, val in zip(self.__fields, data)}
            node = Node(data, self.project)
            self.__items[node.node_id] = node
            return node

        raise ValueError(f"Node {node_id} does not exist in the model")

    def refresh_fields(self) -> None:
        """After adding a field one needs to refresh all the fields recognized by the software"""
        tl = TableLoader()
        tl.load_structure(self._curr, "nodes")
        self.sql = tl.sql
        self.__fields = deepcopy(tl.fields)

    def refresh(self):
        """Refreshes all the nodes in memory"""
        lst = list(self.__items.keys())
        for k in lst:
            del self.__items[k]

    def new_centroid(self, node_id: int) -> Node:
        """Creates a new centroid with a given ID

        :Arguments:
            **node_id** (:obj:`int`): Id of the centroid to be created
        """

        self._curr.execute("select count(*) from nodes where node_id=?", [node_id])
        if self._curr.fetchone()[0] > 0:
            raise Exception("Node_id already exists. Failed to create it")

        data = {key: None for key in self.__fields}
        data["node_id"] = node_id
        data["is_centroid"] = 1
        node = Node(data, self.project)
        self.__items[node_id] = node
        return node

    def save(self):
        for item in self.__items.values():
            item.save()

    @property
    def data(self) -> pd.DataFrame:
        """Returns all nodes data as a Pandas DataFrame

        :Returns:
            **table** (:obj:`DataFrame`): Pandas DataFrame with all the nodes, complete with Geometry
        """
        dl = DataLoader(self.conn, "nodes")
        return dl.load_table()

    def __del__(self):
        self.__items.clear()



from pathlib import Path

import pytest
from aequilibrae.project.field_editor import FieldEditor


class TestFieldEditor:
    @pytest.fixture(scope="class")
    def project_session(self, create_empty_project_session):
        return create_empty_project_session()

    @pytest.fixture(scope="class")
    def database_backup(self, project_session):
        return Path(project_session.project_base_path).joinpath("project_database.sqlite").read_bytes()

    @pytest.fixture
    def cleanup_database(self, project_session, database_backup):
        """An optional fixture that can be used to revert the project's database in case the test
        modifies it. This adds some extra overhead to the test (teardown)
        """
        yield
        Path(project_session.project_base_path).joinpath("project_database.sqlite").write_bytes(database_backup)

    @pytest.fixture(params=["link_types", "links", "modes", "nodes"])
    def table_name(self, request):
        return request.param

    @pytest.fixture
    def table(self, project_session, table_name):
        return FieldEditor(project_session, table_name)

    @pytest.fixture
    def field_name(self, table):
        return next(iter(table._original_values.keys()))

    @pytest.fixture
    def attribute_count(self, table, table_name):
        qry = f'select count(*) from "attributes_documentation" where name_table="{table_name}"'
        return table.project.conn.execute(qry).fetchone()[0]

    def test_building(self, table, attribute_count):
        assert attribute_count == len(table._original_values), "Meta table populated with the wrong number of elements"

    def test_error_when_adding_existing_attribute(self, table, field_name):
        with pytest.raises(ValueError, match="attribute_name already exists"):
            table.add(field_name, "some_value")

    @pytest.mark.parametrize(
        "attribute_name, error",
        [
            ("with space", 'attribute_name can only contain letters, numbers and "_"'),
            ("0starts_with_digit", "attribute_name cannot begin with a digit"),
        ],
    )
    def test_add_invalid_attribute(self, attribute_name, error, table):
        with pytest.raises(ValueError, match=error):
            table.add(attribute_name, "some description")

    @pytest.mark.usefixtures("cleanup_database")
    def test_add_valid_attribute(self, table, table_name, attribute_count):
        project = table.project
        new_attribute = "new_attribute"
        table.add(new_attribute, "some description")
        project.conn.commit()
        curr = project.conn.cursor()
        curr.execute(f'select count(*) from "attributes_documentation" where name_table="{table_name}"')
        q2 = curr.fetchone()[0]
        assert q2 == attribute_count + 1, "Adding element did not work"

        result = curr.execute(
            f'select "{new_attribute}" from "attributes_documentation" where name_table="{table_name}"'
        ).fetchone()[0]
        assert result == new_attribute

    # Here we override the `table_name` fixture. This fixture is not directly used in the test but
    # given as input for the `table` fixture, which we do consume here. This way we change the
    # `table_name` parametrization from four cases to just two, but only for this test.
    @pytest.mark.parametrize(
        "table_name, attribute, description",
        [
            ("link_types", "lanes", "Default number of lanes in each direction. E.g. 2"),
            ("nodes", "is_centroid", "Flag identifying centroids"),
        ],
    )
    def test_retrieve_existing_field(self, table, attribute, description):
        assert attribute in table._original_values
        assert getattr(table, attribute) == description

    @pytest.mark.parametrize(
        "table_name, attribute",
        [
            ("links", "link_id"),
            ("nodes", "node_id"),
        ],
    )
    def test_save(self, table_name, attribute, project_session, table):
        random_val = "some_value"
        setattr(table, attribute, random_val)
        table.save()
        table2 = FieldEditor(project_session, table_name)
        assert getattr(table2, attribute) == random_val, "Did not save values properly"



from copy import deepcopy
import os
from tempfile import gettempdir
from uuid import uuid4
from os.path import join
import sys
from unittest import TestCase

from aequilibrae.paths import Graph
from aequilibrae.paths.results import PathResults
from aequilibrae.utils.create_example import create_example
import numpy as np

# Adds the folder with the data to the path and collects the paths to the files
lib_path = os.path.abspath(os.path.join("..", "../tests"))
sys.path.append(lib_path)

origin = 5
dest = 13


class TestPathResultsDisconnected(TestCase):
    def setUp(self) -> None:
        self.project = create_example(join(gettempdir(), "test_path_disconnected" + uuid4().hex))

    def tearDown(self) -> None:
        self.project.close()

    def test_path_disconnected_delete_link(self):
        self.project.conn.executemany("delete from Links where link_id=?", [[2], [4], [5], [14]])
        self.project.conn.commit()

        self.project.network.build_graphs()
        self.g = self.project.network.graphs["c"]  # type: Graph
        self.g.set_graph("free_flow_time")
        self.g.set_blocked_centroid_flows(False)
        self.r = PathResults()
        self.r.prepare(self.g)
        self.r.compute_path(1, 5)
        self.assertEqual(None, self.r.path, "Failed to return None for disconnected")
        self.r.compute_path(1, 2)
        self.assertEqual(len(self.r.path), 1, "Returned the wrong thing for existing path on disconnected network")

    def test_path_disconnected_penalize_link_in_memory(self):
        links = [2, 4, 5, 14]

        self.project.network.build_graphs()
        g = self.project.network.graphs["c"]  # type: Graph
        g.exclude_links(links)
        g.set_graph("free_flow_time")
        g.set_blocked_centroid_flows(False)
        r = PathResults()
        r.prepare(g)
        r.compute_path(1, 5)
        self.assertEqual(None, r.path, "Failed to return None for disconnected")
        r.compute_path(1, 2)
        self.assertEqual(len(r.path), 1, "Returned the wrong thing for existing path on disconnected network")



from unittest import TestCase
from aequilibrae.paths.AoN import inrets, delta_inrets
from multiprocessing import cpu_count
import numpy as np


class TestInrets(TestCase):
    def test_inrets_funtion(self):
        cores = cpu_count()

        alpha = np.zeros(11)
        beta = np.zeros(11)
        fftime = np.ones(11)
        capacity = np.ones(11)
        congested_times = np.zeros(11)
        delta = np.zeros(11)

        alpha.fill(0.95)
        link_flows = np.arange(11).astype(float) * 0.2

        inrets(congested_times, link_flows, capacity, fftime, alpha, beta, cores)

        should_be = np.array(
            [
                1,
                1.011111111,
                1.028571429,
                1.06,
                1.133333333,
                1.5,
                2.16,
                2.94,
                3.84,
                4.86,
                6,
            ]
        )

        for i in range(11):
            self.assertAlmostEqual(should_be[i], congested_times[i], 5, "Inrets is wrong")

        # Let's check the derivative for sections of the curve
        dx = 0.00000001
        for i in range(1, 20):
            link_flows.fill(1 * 0.1001 * i)

            link_flows += np.arange(11) * dx
            print(link_flows)
            inrets(congested_times, link_flows, capacity, fftime, alpha, beta, cores)
            delta_inrets(delta, link_flows, capacity, fftime, alpha, beta, cores)

            # The derivative needs to be monotonically increasing.
            # np.testing.asser
            self.assertGreater(min(delta[1:] - delta[:-1]), 0, "Delta is not increasing as it should")

            # We check if the analytical solution matches the numerical differentiation
            for j in range(10):
                dydx = (congested_times[j + 1] - congested_times[j]) / dx
                self.assertAlmostEqual(dydx, delta[j + 1], 6, "Problems with derivative for the inrets vdf")
            print(j)



from aequilibrae.project import Project
from aequilibrae.transit import Transit
from aequilibrae.project.database_connection import database_connection
import pytest


@pytest.fixture
def create_project(project: Project):
    Transit(project)


@pytest.mark.parametrize(
    "table, exp_column",
    [
        ("agencies", ["agency_id", "agency", "feed_date", "service_date", "description"]),
        (
            "fare_attributes",
            ["fare_id", "fare", "agency_id", "price", "currency", "payment_method", "transfer", "transfer_duration"],
        ),
        ("fare_rules", ["fare_id", "route_id", "origin", "destination", "contains"]),
        ("fare_zones", ["fare_zone_id", "transit_zone", "agency_id"]),
        ("pattern_mapping", ["pattern_id", "seq", "link", "dir", "geometry"]),
        (
            "routes",
            [
                "pattern_id",
                "route_id",
                "route",
                "agency_id",
                "shortname",
                "longname",
                "description",
                "route_type",
                "seated_capacity",
                "total_capacity",
                "geometry",
            ],
        ),
        ("route_links", ["transit_link", "pattern_id", "seq", "from_stop", "to_stop", "distance", "geometry"]),
        (
            "stops",
            [
                "stop_id",
                "stop",
                "agency_id",
                "link",
                "dir",
                "name",
                "parent_station",
                "description",
                "street",
                "fare_zone_id",
                "transit_zone",
                "route_type",
                "geometry",
            ],
        ),
        ("trips", ["trip_id", "trip", "dir", "pattern_id"]),
        ("trips_schedule", ["trip_id", "seq", "arrival", "departure"]),
    ],
)
def test_create_table(table: str, exp_column: list, create_project):
    curr = database_connection(db_type="transit").cursor()
    curr.execute(f"PRAGMA table_info({table});")
    fields = curr.fetchall()
    fields = [x[1] for x in fields]

    assert exp_column == fields, f"Table {table.upper()} was not created correctly"



from unittest import TestCase
import os
import tempfile
import numpy as np
from aequilibrae.paths import Graph
from os.path import join
from uuid import uuid4
from .parameters_test import centroids
from aequilibrae.project import Project
from ...data import siouxfalls_project
from aequilibrae.paths.results import PathResults

# Adds the folder with the data to the path and collects the paths to the files
# lib_path = os.path.abspath(os.path.join('..', '../tests'))
# sys.path.append(lib_path)
from ...data import path_test, test_graph, test_network
from shutil import copytree, rmtree


class TestGraph(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(tempfile.gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid4().hex)
        copytree(siouxfalls_project, self.temp_proj_folder)
        self.project = Project()
        self.project.open(self.temp_proj_folder)
        self.project.network.build_graphs()
        self.graph = self.project.network.graphs["c"]

    def tearDown(self) -> None:
        self.project.close()

    def test_prepare_graph(self):
        graph = self.project.network.graphs["c"]
        graph.prepare_graph(np.arange(5) + 1)

    def test_set_graph(self):
        self.graph.set_graph(cost_field="distance")
        self.graph.set_blocked_centroid_flows(block_centroid_flows=True)
        self.assertEqual(self.graph.num_zones, 24, "Number of centroids not properly set")
        self.assertEqual(self.graph.num_links, 76, "Number of links not properly set")
        self.assertEqual(self.graph.num_nodes, 24, "Number of nodes not properly set - " + str(self.graph.num_nodes))

    def test_save_to_disk(self):
        self.graph.save_to_disk(join(path_test, "aequilibrae_test_graph.aeg"))
        self.graph_id = self.graph.__id__

    def test_load_from_disk(self):
        self.test_save_to_disk()
        reference_graph = Graph()
        reference_graph.load_from_disk(test_graph)

        new_graph = Graph()
        new_graph.load_from_disk(join(path_test, "aequilibrae_test_graph.aeg"))

    def test_available_skims(self):
        self.graph.prepare_graph(np.arange(5) + 1)
        avail = self.graph.available_skims()
        data_fields = [
            "distance",
            "name",
            "lanes",
            "capacity",
            "speed",
            "b",
            "free_flow_time",
            "power",
            "colum",
            "volume",
            "modes",
        ]
        for i in data_fields:
            if i not in avail:
                self.fail("Skim availability with problems")

    def test_exclude_links(self):
        # excludes a link before any setting or preparation
        self.graph.set_blocked_centroid_flows(False)

        self.graph.set_graph("distance")
        r1 = PathResults()
        r1.prepare(self.graph)
        r1.compute_path(20, 21)
        self.assertEqual(list(r1.path), [62])

        r1 = PathResults()
        self.graph.exclude_links([62])
        r1.prepare(self.graph)
        r1.compute_path(20, 21)
        self.assertEqual(list(r1.path), [63, 69])



from collections import OrderedDict


class PathStorage:
    """
    This class is designed to store path-computation objects to take advantage of the fact that AequilibraE
    preserves the entire shortest path tree when computing a path between two nodes and can re-trace the
    same tree for a path from the same origin to a different destination.

    Since this caching in memory can take too much memory, the *threshold* parameter exists to limit the number
    of path objects kept in memory.

    If you have a large amount of memory in your system, you can set the threshold class variable accordingly.
    """

    def __init__(self):
        self.graphs = {}
        self.storage = OrderedDict()
        self.uses = 0
        self.total_paths = 0
        self.threshold = 50

    def add_graph(self, graph, mode_id):
        if mode_id in self.graphs:
            return

        self.graphs[mode_id] = graph
        self.storage[mode_id] = OrderedDict()

    def get_path_results(self, origin, mode_id):
        from aequilibrae.paths import PathResults

        self.uses += 1
        if origin in self.storage[mode_id]:
            # We move the last used element from the list to the most recent position
            self.storage[mode_id].move_to_end(origin, last=True)
            return self.storage[mode_id][origin]

        # We check if our cache is getting too large
        stored_values = [len(data) for data in self.storage.values()]
        if sum(stored_values) > self.threshold:
            # If we reached the maximum number of paths to store, we remove the element that has not been used
            # in the longest time from the mode ID with the biggest cache
            mode_to_clean = list(self.storage.keys())[stored_values.index(max(stored_values))]
            self.storage[mode_to_clean].clear()

        graph = self.graphs[mode_id]
        res = PathResults()
        res.prepare(graph)
        d = graph.centroids[0] if origin != graph.centroids[0] else graph.centroids[1]
        res.compute_path(origin, d)
        self.storage[mode_id][origin] = res
        self.total_paths += 1

        return res

    def clear(self):
        self.storage.clear()
        self.graphs.clear()
        self.uses = 0
        self.total_paths = 0



import random
import sqlite3
import string

import pytest
from aequilibrae import Project


class TestAbout:
    @pytest.fixture
    def project(self, create_project):
        return create_project()

    def test_create_and_list(self, project: Project):
        project.about.create()
        list = project.about.list_fields()
        expected = [
            "model_name",
            "region",
            "description",
            "author",
            "license",
            "scenario_name",
            "year",
            "scenario_description",
            "model_version",
            "project_id",
            "aequilibrae_version",
            "projection",
        ]
        assert not set(list) ^ set(expected), "About table does not have all expected fields"

    def test_warning_when_creating_twice(self, project: Project):
        project.about.create()
        project.about.create()
        last_log = project.log().contents()[-1]
        assert "About table already exists" in last_log

    # idea from https://stackoverflow.com/a/2030081/1480643
    def randomword(self, length):
        letters = string.ascii_lowercase + "_"
        return "".join(random.choice(letters) for i in range(length))

    def test_add_info_field(self, project: Project):
        project.about.create()

        all_added = set()
        for t in range(30):
            k = self.randomword(random.randint(1, 15))
            if k not in all_added:
                all_added.add(k)
                project.about.add_info_field(k)

        curr = project.conn.cursor()
        curr.execute("select infoname from 'about'")

        charac = [x[0] for x in curr.fetchall()]
        for k in all_added:
            if k not in charac:
                self.fail(f"Failed to add {k}")

        # Should fail when trying to add a repeated guy
        with pytest.raises(sqlite3.IntegrityError):
            project.about.add_info_field("description")

        # Should fail when trying to add a repeated guy
        with pytest.raises(ValueError):
            project.about.add_info_field("descr1ption")

    def test_write_back(self, project: Project):
        base_path = project.project_base_path
        project.about.create()
        project.about.add_info_field("good_info_field_perhaps")

        val = self.randomword(random.randint(1, 15))
        project.about.good_info_field_perhaps = val

        val2 = self.randomword(random.randint(30, 250))
        project.about.description = val2

        project.about.write_back()

        project.close()
        del project

        project = Project()
        project.open(base_path)
        assert val == project.about.good_info_field_perhaps, "failed to save data to about table"
        assert val2 == project.about.description, "failed to save data to about table"



from unittest import TestCase
import tempfile
import os
from random import choice
from warnings import warn
from shutil import copytree, rmtree
from aequilibrae.project import Project
import uuid
import sqlite3
from aequilibrae import logger
from ...data import no_triggers_project


class TestProject(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(tempfile.gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        copytree(no_triggers_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.curr = self.proj.conn.cursor()

        # Modes to add
        sql = "INSERT INTO modes (mode_name, mode_id) VALUES (?, ?);"
        for mid in ["p", "l", "g", "x", "y", "d", "k", "a", "r", "n", "m"]:
            self.curr.execute(sql, [f"mode_{mid}", mid])

        self.proj.conn.commit()

        curr = self.proj.conn.cursor()
        self.rtree = True
        try:
            curr.execute("SELECT rtreecheck('idx_nodes_geometry');")
        except Exception as e:
            self.rtree = False
            warn(f"RTREE not available --> {e.args}")

        root = os.path.dirname(os.path.realpath(__file__)).replace("tests", "")
        qry_file = os.path.join(root, "database_specification/network/triggers/modes_table_triggers.sql")
        with open(qry_file, "r") as sql_file:
            self.queries = sql_file.read()
        self.queries = [cmd for cmd in self.queries.split("#")]

    def tearDown(self) -> None:
        self.proj.close()

    def __get_query(self, qry: str) -> str:
        for query in self.queries:
            if qry in query:
                return query
        raise FileNotFoundError("QUERY DOES NOT EXIST")

    def test_all_tests_considered(self):
        tests_added = list(self.__dir__())
        tests_added = [x[5:] for x in tests_added if x[:5] == "test_"]

        for trigger in self.queries:
            if "TRIGGER" in trigger.upper():
                found = [x for x in tests_added if x in trigger]
                if not found:
                    self.fail(f"Trigger not tested. {trigger}")

    def test_mode_single_letter_update(self):
        cmd = self.__get_query("mode_single_letter_update")

        sql = "UPDATE 'modes' SET mode_id= 'ttt' where mode_id='b'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        with self.assertRaises(sqlite3.IntegrityError):
            sql = "UPDATE 'modes' SET mode_id= 'gg' where mode_id='w'"
            self.curr.execute(sql)

    def test_mode_single_letter_insert(self):
        cmd = self.__get_query("mode_single_letter_insert")

        sql = "INSERT INTO 'modes' (mode_name, mode_id) VALUES(?, ?)"
        self.curr.execute(sql, ["testasdasd", "pp"])

        self.curr.execute(cmd)

        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql, ["test1b", "mm"])

    def test_mode_keep_if_in_use_updating(self):
        cmd = self.__get_query("mode_keep_if_in_use_updating")

        sql = "UPDATE 'modes' SET mode_id= 'h' where mode_id='g'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "UPDATE 'modes' SET mode_id= 'j' where mode_id='l'"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_mode_keep_if_in_use_deleting(self):
        cmd = self.__get_query("mode_keep_if_in_use_deleting")

        sql = "DELETE FROM 'modes' where mode_id='p'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "DELETE FROM 'modes' where mode_id='c'"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_modes_on_links_update(self):
        cmd = self.__get_query("modes_on_links_update")

        sql = "UPDATE 'links' SET modes= 'qwerty' where link_id=55"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "UPDATE 'links' SET modes= 'azerty' where link_id=56"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_modes_length_on_links_update(self):
        cmd = self.__get_query("modes_length_on_links_update")

        sql = "UPDATE 'links' SET modes= '' where modes='wb'"
        self.curr.execute(sql)

        self.curr.execute(cmd)

        sql = "UPDATE 'links' SET modes= '' where modes='bw'"
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute(sql)

    def test_modes_on_nodes_table_update_a_node(self):
        cmd = self.__get_query("modes_on_nodes_table_update_a_node")

        sql = "UPDATE 'links' SET a_node= 1 where a_node=3"
        self.curr.execute(sql)

        sql = "SELECT modes from nodes where node_id=1"
        self.curr.execute(sql)
        i = self.curr.fetchone()[0]
        self.assertEqual(i, "ct")

        self.curr.execute(cmd)

        k = ""
        for n in [2, 5]:
            for f in ["a_node", "b_node"]:
                self.curr.execute(f"SELECT modes from links where {f}={n}")
                k += self.curr.fetchone()[0]

        existing = set(k)

        sql = "UPDATE 'links' SET a_node= 2 where a_node=5"
        self.curr.execute(sql)

        sql = "SELECT modes from nodes where node_id=2"
        self.curr.execute(sql)
        i = set(self.curr.fetchone()[0])
        self.assertEqual(i, existing)

    def test_modes_on_nodes_table_update_b_node(self):
        cmd = self.__get_query("modes_on_nodes_table_update_b_node")
        sql = "UPDATE 'links' SET b_node= 1 where b_node=3"
        self.curr.execute(sql)

        sql = "SELECT modes from nodes where node_id=1"
        self.curr.execute(sql)
        i = self.curr.fetchone()[0]
        self.assertEqual(i, "ct")

        self.curr.execute(cmd)

        sql = "UPDATE 'links' SET b_node= 2 where b_node=4"
        self.curr.execute(sql)

        sql = "SELECT modes from nodes where node_id=2"
        self.curr.execute(sql)
        i = self.curr.fetchone()[0]
        self.assertEqual(i, "ctw")

    def test_modes_on_nodes_table_update_links_modes(self):
        cmd = self.__get_query("modes_on_nodes_table_update_links_modes")
        sql = "UPDATE 'links' SET modes= 'x' where a_node=24"
        self.curr.execute(sql)

        sql = "SELECT modes from nodes where node_id=24"
        self.curr.execute(sql)
        i = self.curr.fetchone()[0]
        self.assertEqual(i, "c")

        self.curr.execute(cmd)

        sql = "UPDATE 'links' SET 'modes'= 'y' where a_node=24"
        self.curr.execute(sql)

        sql = "SELECT modes from nodes where node_id=24"
        self.curr.execute(sql)
        i = self.curr.fetchone()[0]
        self.assertIn("c", i)
        self.assertIn("y", i)

        sql = "UPDATE 'links' SET 'modes'= 'r' where b_node=24"
        self.curr.execute(sql)

        sql = "SELECT modes from nodes where node_id=24"
        self.curr.execute(sql)
        i = self.curr.fetchone()[0]
        self.assertIn("r", i)
        self.assertIn("y", i)

    def test_modes_on_links_insert(self):
        cmd = self.__get_query("modes_on_links_insert")
        if self.rtree:
            self.curr.execute("pragma table_info(links)")
            f = self.curr.fetchall()
            fields = {x[1]: x[0] for x in f}

            sql = "select * from links where link_id=10"
            self.curr.execute(sql)
            a = [x for x in self.curr.fetchone()]
            a[fields["modes"]] = "as12"
            a[fields["link_id"]] = 1234
            a[fields["a_node"]] = 999
            a[fields["b_node"]] = 888
            a[0] = 1234

            idx = ",".join(["?"] * len(a))
            self.curr.execute(f"insert into links values ({idx})", a)
            self.curr.execute("delete from links where link_id=1234")

            self.curr.execute(cmd)

            with self.assertRaises(sqlite3.IntegrityError):
                self.curr.execute(f"insert into links values ({idx})", a)

    def test_modes_length_on_links_insert(self):
        cmd = self.__get_query("modes_length_on_links_insert")
        if self.rtree:
            self.curr.execute("pragma table_info(links)")
            f = self.curr.fetchall()
            fields = {x[1]: x[0] for x in f}

            sql = "select * from links where link_id=70"
            self.curr.execute(sql)
            a = [x for x in self.curr.fetchone()]
            a[fields["modes"]] = ""
            a[fields["link_id"]] = 4321
            a[fields["a_node"]] = 888
            a[fields["b_node"]] = 999
            a[0] = 4321

            idx = ",".join(["?"] * len(a))
            self.curr.execute(f"insert into links values ({idx})", a)
            self.curr.execute("delete from links where link_id=4321")

            self.curr.execute(cmd)

            with self.assertRaises(sqlite3.IntegrityError):
                self.curr.execute(f"insert into links values ({idx})", a)

    def test_keep_at_least_one(self):
        cmd = self.__get_query("mode_keep_at_least_one")

        self.curr.execute("Delete from modes;")
        self.curr.execute("select count(*) from modes;")
        self.assertEqual(self.curr.fetchone()[0], 0, "We were not able to clear the m odes table as expected")

        self.curr.execute('insert into modes(mode_id, mode_name) VALUES("k", "some_name")')
        self.curr.execute(cmd)
        with self.assertRaises(sqlite3.IntegrityError):
            self.curr.execute("Delete from modes;")

    def test_modes_on_nodes_table_update_nodes_modes(self):
        cmd = self.__get_query("modes_on_nodes_table_update_nodes_modes")
        self.curr.execute("select node_id, modes from nodes where length(modes)>0")
        dt = self.curr.fetchall()

        x = choice(dt)

        self.curr.execute(f'update nodes set modes="abcdefgq" where node_id={x[0]}')
        self.curr.execute(f"select node_id, modes from nodes where node_id={x[0]}")
        z = self.curr.fetchone()
        if z == x:
            self.fail("Modes field on nodes layer is being preserved by unknown mechanism")

        self.curr.execute(cmd)

        y = choice(dt)
        while y == x:
            y = choice(dt)

        # We try to force the change to make sure it was correctly filled to begin with
        self.curr.execute(f'update nodes set modes="hgfedcba" where node_id={y[0]}')

        self.curr.execute(f"select node_id, modes from nodes where node_id={y[0]}")
        k = self.curr.fetchone()

        self.curr.execute(f'update nodes set modes="abcdefgq" where node_id={y[0]}')

        self.curr.execute(f"select node_id, modes from nodes where node_id={y[0]}")
        z = self.curr.fetchone()

        self.assertEqual(z, k, "Failed to preserve the information on modes for the nodes")



from unittest import TestCase
from tempfile import gettempdir
from shutil import copytree
import os
import uuid
import pandas as pd
from aequilibrae.project import Project
from aequilibrae.parameters import Parameters
from warnings import warn
import random
from ...data import siouxfalls_project


class TestNetwork(TestCase):
    def test_create_from_gmns(self):
        proj_path = os.path.join(gettempdir(), uuid.uuid4().hex)
        self.project = Project()
        self.project.new(proj_path)

        link_file = "https://raw.githubusercontent.com/zephyr-data-specs/GMNS/development/Small_Network_Examples/Arlington_Signals/link.csv"
        node_file = "https://raw.githubusercontent.com/zephyr-data-specs/GMNS/development/Small_Network_Examples/Arlington_Signals/node.csv"
        use_group_file = "https://raw.githubusercontent.com/zephyr-data-specs/GMNS/development/Small_Network_Examples/Arlington_Signals/use_group.csv"

        new_link_fields = {
            "bridge": {"description": "bridge flag", "type": "text", "required": False},
            "tunnel": {"description": "tunnel flag", "type": "text", "required": False},
        }
        new_node_fields = {
            "port": {"description": "port flag", "type": "text", "required": False},
            "hospital": {"description": "hoospital flag", "type": "text", "required": False},
        }
        par = Parameters()
        par.parameters["network"]["gmns"]["link"]["fields"].update(new_link_fields)
        par.parameters["network"]["gmns"]["node"]["fields"].update(new_node_fields)
        par.write_back()

        self.project.network.create_from_gmns(
            link_file_path=link_file, node_file_path=node_file, use_group_path=use_group_file, srid=32619
        )

        gmns_node_df = pd.read_csv(node_file)
        gmns_link_df = pd.read_csv(link_file)

        curr = self.project.conn.cursor()
        curr.execute("""select count(*) from nodes""")
        nd_ct = curr.fetchone()[0]

        if nd_ct != gmns_node_df.shape[0]:
            warn("Number of nodes created is different than expected.")
            return

        rand_lk = random.choice([x[0] for x in curr.execute("""select link_id from links""").fetchall()])
        from_node = gmns_link_df.loc[gmns_link_df.link_id == rand_lk, "from_node_id"].item()
        to_node = gmns_link_df.loc[gmns_link_df.link_id == rand_lk, "to_node_id"].item()
        a_node = curr.execute(f"""select a_node from links where link_id = {rand_lk}""").fetchone()[0]
        b_node = curr.execute(f"""select b_node from links where link_id = {rand_lk}""").fetchone()[0]

        if from_node != a_node or to_node != b_node:
            self.fail("At least one link is disconnected from its start/end nodes")

    def test_export_to_gmns(self):
        output_path = os.path.join(gettempdir(), uuid.uuid4().hex)
        if not os.path.exists(output_path):
            os.mkdir(output_path)

        self.temp_proj_folder = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, self.temp_proj_folder)
        self.project = Project()
        self.project.open(self.temp_proj_folder)

        self.project.network.export_to_gmns(output_path)



from math import floor
import string
from unittest import TestCase
from random import choice, randint
from shutil import copyfile
import os
from os.path import join
from shutil import copytree
import uuid
from tempfile import gettempdir
from ...data import siouxfalls_project
from aequilibrae.project import Project


class TestMatrices(TestCase):
    def setUp(self) -> None:
        proj_dir = os.path.join(gettempdir(), uuid.uuid4().hex)
        copytree(siouxfalls_project, proj_dir)

        self.project = Project()
        self.project.open(proj_dir)
        self.matrices = self.project.matrices
        self.matrices.reload()
        self.curr = self.project.conn.cursor()

    def tearDown(self) -> None:
        self.project.close()
        del self.project
        del self.matrices

    def randomword(self, length):
        allowed_characters = string.ascii_letters + '_123456789@!()[]{};:"    -'
        val = "".join(choice(allowed_characters) for i in range(length))
        if val[0] == "_" or val[-1] == "_":
            return self.randomword(length)
        return val

    def test_set_record(self):
        rec = self.matrices.get_record("omx2")
        with self.assertRaises(ValueError):
            rec.name = "omx"

        with self.assertRaises(ValueError):
            rec.file_name = "sfalls_skims.omx"

        rec.file_name = "SiouxFalls.omx"
        self.assertEqual(rec.cores, 1, "Setting a file that exists did not correct the number of cores")

    def test_clear_database(self):
        self.__mat_count(3, "The test data started wrong")

        self.matrices.clear_database()
        self.__mat_count(2, "Did not clear the database appropriately")

    def test_update_database(self):
        self.__mat_count(3, "The test data started wrong")

        self.matrices.update_database()
        self.__mat_count(4, "Did not add to the database appropriately")

        rec = self.matrices.get_record("omx")
        existing = join(rec.fldr, rec.file_name)
        new_name = "test_name.omx"
        new_name1 = "test_name1.omx"

        copyfile(existing, join(rec.fldr, new_name))
        record = self.matrices.new_record("test_name1.omx", new_name)
        record.save()

        copyfile(existing, join(rec.fldr, new_name1))
        self.matrices.update_database()

    def test_get_matrix(self):
        with self.assertRaises(Exception):
            _ = self.matrices.get_matrix("omxq")

        mat = self.matrices.get_matrix("oMx")
        mat.computational_view()
        self.assertEqual(floor(mat.matrix_view.sum()), 20309, "Matrix loaded incorrectly")

    def test_get_record(self):
        rec = self.matrices.get_record("omx")
        self.assertEqual(rec.cores, 4, "record populated wrong. Number of cores")
        self.assertEqual(rec.description, None, "record populated wrong. Description")

    def test_record_update_cores(self):
        rec = self.matrices.get_record("omx")
        rec.update_cores()
        self.assertEqual(rec.cores, 2, "Cores update did not work")

    def test_save_record(self):
        rec = self.matrices.get_record("omx")

        text = self.randomword(randint(30, 100))
        rec.description = text
        rec.save()

        self.curr.execute('select description from matrices where name="omx";')
        self.assertEqual(text, self.curr.fetchone()[0], "Saving matrix record description failed")

    def test_delete(self):
        self.matrices.delete_record("omx")
        self.curr.execute('select count(*) from matrices where name="omx";')

        self.assertEqual(0, self.curr.fetchone()[0], " Deleting matrix record failed")

        with self.assertRaises(Exception):
            self.matrices.get_record("omx")

    def test_list(self):
        df = self.matrices.list()
        self.curr.execute("select count(*) from Matrices")
        self.assertEqual(df.shape[0], self.curr.fetchone()[0], "Returned the wrong number of matrices in the database")

        self.assertEqual(df[df.status == "file missing"].shape[0], 1, "Wrong # of records for missing matrix files")

    def __mat_count(self, should_have: int, error_message: str) -> None:
        self.curr.execute("Select count(*) from Matrices;")
        self.assertEqual(self.curr.fetchone()[0], should_have, error_message)



import os
from tempfile import gettempdir
from unittest import TestCase
import uuid
from aequilibrae.utils.create_example import create_example
from aequilibrae.paths import Graph
from aequilibrae.paths.results import AssignmentResults
from aequilibrae.paths.all_or_nothing import allOrNothing
from ...data import test_graph


# TODO: Add checks for results for this test (Assignment AoN)
class TestAllOrNothing(TestCase):
    def setUp(self) -> None:
        proj_path = os.path.join(gettempdir(), "test_traffic_assignment_" + uuid.uuid4().hex)
        self.project = create_example(proj_path)
        self.project.network.build_graphs()
        self.g = self.project.network.graphs["c"]  # type: Graph
        self.g.set_graph("distance")
        self.g.set_skimming("distance")
        self.g.set_blocked_centroid_flows(False)

        self.matrix = self.project.matrices.get_matrix("demand_aem")
        self.matrix.computational_view()
        self.matrix2 = self.project.matrices.get_matrix("demand_omx")
        self.matrix2.computational_view()
        self.matrix2.matrix_view *= 2

    def tearDown(self) -> None:
        self.matrix.close()
        self.matrix2.close()
        self.project.close()

    def test_skimming_on_assignment(self):
        res = AssignmentResults()

        res.prepare(self.g, self.matrix)

        self.g.set_skimming([])
        self.g.set_blocked_centroid_flows(True)
        assig = allOrNothing(self.matrix, self.g, res)
        assig.execute()

        if res.skims.distance.sum() > 0:
            self.fail("skimming for nothing during assignment returned something different than zero")

        res.prepare(self.g, self.matrix)

        assig = allOrNothing(self.matrix, self.g, res)
        assig.execute()

    def test_execute(self):
        # Loads and prepares the graph
        res1 = AssignmentResults()
        res1.prepare(self.g, self.matrix)
        assig1 = allOrNothing(self.matrix, self.g, res1)
        assig1.execute()

        res2 = AssignmentResults()
        res2.prepare(self.g, self.matrix2)
        assig2 = allOrNothing(self.matrix2, self.g, res2)
        assig2.execute()

        load1 = res1.get_load_results()
        load2 = res2.get_load_results()

        self.assertEqual(list(load1.matrix_tot * 2), list(load2.matrix_tot), "Something wrong with the AoN")



# The conftest.py file serves as a means of providing fixtures for an entire directory.
# Fixtures defined in a conftest.py can be used by any test in that package without
# needing to import them (pytest will automatically discover them).

import os
import uuid
from shutil import copytree

import pytest

from aequilibrae import Project
from aequilibrae.project.database_connection import database_connection
from aequilibrae.transit import Transit
from aequilibrae.utils.create_example import create_example
from aequilibrae.utils.spatialite_utils import ensure_spatialite_binaries

from ..data import siouxfalls_project

DEFAULT_PROJECT = siouxfalls_project
ensure_spatialite_binaries()


def project_factory_fixture(scope):
    @pytest.fixture(scope=scope)
    def create_project_fixture(tmp_path_factory):
        base_dir = tmp_path_factory.mktemp(f"projects_{scope}")
        projects = []

        def _create_project(name=None, source_dir=DEFAULT_PROJECT):
            proj_dir = base_dir / (name or uuid.uuid4().hex)
            copytree(source_dir, proj_dir)
            project = Project()
            project.open(str(proj_dir))
            projects.append(project)
            return project

        yield _create_project

        for project in projects:
            project.close()

    return create_project_fixture


create_project = project_factory_fixture(scope="function")
create_project_session = project_factory_fixture(scope="session")


@pytest.fixture
def create_empty_project(_empty_project, create_project):
    def _create_empty_project(name=None):
        return create_project(name=name, source_dir=_empty_project)

    return _create_empty_project


@pytest.fixture(scope="session")
def create_empty_project_session(_empty_project, create_project_session):
    def _create_empty_project(name=None):
        return create_project_session(name=name, source_dir=_empty_project)

    return _create_empty_project


# This fixture creates a default empty structure on disk that can be used as the
# source folder for creating temporary empty projects
@pytest.fixture(scope="session")
def _empty_project(tmp_path_factory):
    proj_dir = tmp_path_factory.mktemp("_empty_project") / uuid.uuid4().hex
    project = Project()
    project.new(str(proj_dir))
    return proj_dir


@pytest.fixture
def project(create_empty_project):
    return create_empty_project()


@pytest.fixture
def create_path(tmp_path):
    return tmp_path / uuid.uuid4().hex


@pytest.fixture
def create_gtfs_project(create_path):
    prj = create_example(create_path, "coquimbo")

    if os.path.isfile(os.path.join(create_path, "public_transport.sqlite")):
        os.remove(os.path.join(create_path, "public_transport.sqlite"))

    data = Transit(prj)

    yield data
    prj.close()


@pytest.fixture
def transit_conn(create_gtfs_project):
    return database_connection("transit")



from aequilibrae.transit import Transit
from aequilibrae.context import activate_project
from aequilibrae.project.database_connection import database_connection
import pytest


class TestDatabaseConnection:
    def test_cannot_connect_when_no_active_project(self):
        activate_project(None)
        with pytest.raises(FileNotFoundError):
            database_connection("network")

    def test_connection_with_new_project(self, project):
        conn = database_connection(db_type="network", project_path=project.project_base_path)
        cursor = conn.cursor()
        cursor.execute("select count(*) from links")
        assert cursor.fetchone()[0] == 0, "Returned more links thant it should have"

    def test_connection_with_transit(self, project):
        Transit(project)
        conn = database_connection(db_type="transit", project_path=project.project_base_path)
        cursor = conn.cursor()
        cursor.execute("select count(*) from routes")
        assert cursor.fetchone()[0] == 0, "Returned more routes thant it should have"



import unittest
from tempfile import gettempdir
from uuid import uuid4
from os.path import join
from aequilibrae.paths.results import SkimResults
from aequilibrae.paths import skimming_single_origin
from aequilibrae.paths.multi_threaded_skimming import MultiThreadedNetworkSkimming
import numpy as np
from aequilibrae.utils.create_example import create_example

# Adds the folder with the data to the path and collects the paths to the files
from ...data import path_test, test_graph


class TestSkimming_single_origin(unittest.TestCase):
    def setUp(self) -> None:
        path = join(gettempdir(), "skim_test_" + uuid4().hex)
        self.project = create_example(path)

    def tearDown(self) -> None:
        self.project.close()

    def test_skimming_single_origin(self):
        self.project.network.build_graphs()
        g = self.project.network.graphs["c"]
        g.set_graph(cost_field="distance")
        g.set_skimming("distance")

        origin = np.random.choice(g.centroids[:-1], 1)[0]

        # skimming results
        res = SkimResults()
        res.prepare(g)
        aux_result = MultiThreadedNetworkSkimming()
        aux_result.prepare(g, res)

        a = skimming_single_origin(origin, g, res, aux_result, 0)
        tot = np.sum(res.skims.distance[origin, :])
        if tot > 10e10:
            self.fail("Skimming was not successful. At least one np.inf returned for origin {}.".format(origin))

        if a != origin:
            self.fail("Skimming returned an error: {} for origin {}".format(a, origin))



from os.path import join
import pickle
import uuid
from datetime import datetime
from typing import List, Tuple
import numpy as np
import pandas as pd
from aequilibrae.context import get_logger
from aequilibrae.paths.AoN import build_compressed_graph


class Graph(object):
    """
    Graph class
    """

    def __init__(self, logger=None):
        self.logger = logger or get_logger()
        self.__integer_type = np.int64
        self.__float_type = np.float64

        self.required_default_fields = ["link_id", "a_node", "b_node", "direction", "id"]
        self.__required_default_types = [
            self.__integer_type,
            self.__integer_type,
            self.__integer_type,
            np.int8,
            self.__integer_type,
        ]
        self.other_fields = ""
        self.mode = ""
        self.date = str(datetime.now())

        self.description = "No description added so far"

        self.num_links = -1
        self.num_nodes = -1
        self.num_zones = -1

        self.compact_num_links = -1
        self.compact_num_nodes = -1

        self.network = pd.DataFrame([])  # This method will hold ALL information on the network
        self.graph = pd.DataFrame([])  # This method will hold an array with ALL fields in the graph.

        self.compact_graph = pd.DataFrame([])  # This method will hold an array with ALL fields in the graph.

        # These are the fields actually used in computing paths
        self.all_nodes = np.array(0)  # Holds an array with all nodes in the original network
        self.nodes_to_indices = np.array(0)  # Holds the reverse of the all_nodes
        self.fs = np.array([])  # This method will hold the forward star for the graph
        self.cost = np.array([])  # This array holds the values being used in the shortest path routine
        self.skims = None

        self.compact_all_nodes = np.array(0)  # Holds an array with all nodes in the original network
        self.compact_nodes_to_indices = np.array(0)  # Holds the reverse of the all_nodes
        self.compact_fs = np.array([])  # This method will hold the forward star for the graph
        self.compact_cost = np.array([])  # This array holds the values being used in the shortest path routine
        self.compact_skims = None

        self.capacity = np.array([])  # Array holds the capacity for links
        self.free_flow_time = np.array([])  # Array holds the free flow travel time by link

        # sake of the Cython code
        self.skim_fields = []  # List of skim fields to be used in computation
        self.cost_field = False  # Name of the cost field

        self.block_centroid_flows = True
        self.penalty_through_centroids = np.inf

        self.centroids = None  # NumPy array of centroid IDs

        self.g_link_crosswalk = np.array([])  # 4 a link ID in the BIG graph, a corresponding link in the compressed 1

        # Randomly generate a unique Graph ID randomly
        self.__id__ = uuid.uuid4().hex

    def default_types(self, tp: str):
        """
        Returns the default integer and float types used for computation

        :Arguments:
            **tp** (:obj:`str`): data type. 'int' or 'float'
        """
        if tp == "int":
            return self.__integer_type
        elif tp == "float":
            return self.__float_type
        else:
            raise ValueError("It must be either a int or a float")

    def prepare_graph(self, centroids: np.ndarray) -> None:
        """
        Prepares the graph for a computation for a certain set of centroids

        Under the hood, if sets all centroids to have IDs from 1 through **n**,
        which should correspond to the index of the matrix being assigned.

        This is what enables having any node IDs as centroids, and it relies on
        the inference that all links connected to these nodes are centroid
        connectors.

        :Arguments:
            **centroids** (:obj:`np.ndarray`): Array with centroid IDs. Mandatory type Int64, unique and positive
        """
        self.__network_error_checking__()

        # Creates the centroids

        if centroids is None or not isinstance(centroids, np.ndarray):
            raise ValueError("Centroids need to be a NumPy array of integers 64 bits")
        if not np.issubdtype(centroids.dtype, np.integer):
            raise ValueError("Centroids need to be a NumPy array of integers 64 bits")
        if centroids.shape[0] == 0:
            raise ValueError("You need at least one centroid")
        if centroids.min() <= 0:
            raise ValueError("Centroid IDs need to be positive")
        if centroids.shape[0] != np.unique(centroids).shape[0]:
            raise ValueError("Centroid IDs are not unique")
        self.centroids = np.array(centroids, np.uint32)

        self.network = self.network.astype(
            {
                "direction": np.int8,
                "a_node": self.__integer_type,
                "b_node": self.__integer_type,
                "link_id": self.__integer_type,
            }
        )

        properties = self.__build_directed_graph(self.network, centroids)
        self.all_nodes, self.num_nodes, self.nodes_to_indices, self.fs, self.graph = properties

        # We generate IDs that we KNOW will be constant across modes
        self.graph.sort_values(by=["link_id", "direction"], inplace=True)
        self.graph["__supernet_id__"] = np.arange(self.graph.shape[0]).astype(self.__integer_type)
        self.graph.sort_values(by=["a_node", "b_node"], inplace=True)

        self.num_links = self.graph.shape[0]
        self.__build_derived_properties()

        self.__build_compressed_graph()
        self.compact_num_links = self.compact_graph.shape[0]

    def __build_compressed_graph(self):
        build_compressed_graph(self)

        # We build a groupby to save time later
        self.__graph_groupby = self.graph.groupby(["__compressed_id__"])

    def __build_directed_graph(self, network: pd.DataFrame, centroids: np.ndarray):
        all_titles = list(network.columns)

        not_pos = network.loc[network.direction != 1, :]
        not_negs = network.loc[network.direction != -1, :]

        names, types = self.__build_column_names(all_titles)
        neg_names = []
        for name in names:
            if name in not_pos.columns:
                neg_names.append(name)
            elif name + "_ba" in not_pos.columns:
                neg_names.append(name + "_ba")
        not_pos = pd.DataFrame(not_pos, copy=True)[neg_names]
        not_pos.columns = names
        not_pos.loc[:, "direction"] = -1
        aux = np.array(not_pos.a_node.values, copy=True)
        not_pos.loc[:, "a_node"] = not_pos.loc[:, "b_node"]
        not_pos.loc[:, "b_node"] = aux[:]
        del aux

        pos_names = []
        for name in names:
            if name in not_negs.columns:
                pos_names.append(name)
            elif name + "_ab" in not_negs.columns:
                pos_names.append(name + "_ab")
        not_negs = pd.DataFrame(not_negs, copy=True)[pos_names]
        not_negs.columns = names
        not_negs.loc[:, "direction"] = 1

        df = pd.concat([not_negs, not_pos])

        # Now we take care of centroids
        nodes = np.unique(np.hstack((df.a_node.values, df.b_node.values))).astype(self.__integer_type)
        nodes = np.setdiff1d(nodes, centroids, assume_unique=True)
        all_nodes = np.hstack((centroids, nodes)).astype(self.__integer_type)

        num_nodes = all_nodes.shape[0]

        nodes_to_indices = np.repeat(-1, int(all_nodes.max()) + 1)
        nlist = np.arange(num_nodes)
        nodes_to_indices[all_nodes] = nlist

        df.loc[:, "a_node"] = nodes_to_indices[df.a_node.values][:]
        df.loc[:, "b_node"] = nodes_to_indices[df.b_node.values][:]
        df = df.sort_values(by=["a_node", "b_node"])
        df.index = np.arange(df.shape[0])
        df["id"] = np.arange(df.shape[0])
        fs = np.empty(num_nodes + 1, dtype=self.__integer_type)
        fs.fill(-1)
        y, x, _ = np.intersect1d(df.a_node.values, nlist, assume_unique=False, return_indices=True)
        fs[y] = x[:]
        fs[-1] = df.shape[0]
        for i in range(num_nodes, 1, -1):
            if fs[i - 1] == -1:
                fs[i - 1] = fs[i]

        nans = ", ".join([i for i in df.columns if df[i].isnull().any().any()])
        if nans:
            self.logger.warning(f"Field(s) {nans} has(ve) at least one NaN value. Check your computations")

        df.loc[:, "b_node"] = df.b_node.values.astype(self.__integer_type)
        df.loc[:, "id"] = df.id.values.astype(self.__integer_type)
        df.loc[:, "link_id"] = df.link_id.values.astype(self.__integer_type)
        df["direction"] = df.direction.values.astype(np.int8)

        return all_nodes, num_nodes, nodes_to_indices, fs, df

    def exclude_links(self, links: list) -> None:
        """
        Excludes a list of links from a graph by setting their B node equal to their A node

        :Arguments:
            **links** (:obj:`list`): List of link IDs to be excluded from the graph
        """
        filter = self.network.link_id.isin(links)
        # We check is the list makes sense in order to warn the user
        if filter.sum() != len(set(links)):
            self.logger.warning("At least one link does not exist in the network and therefore cannot be excluded")

        self.network.loc[filter, "b_node"] = self.network.loc[filter, "a_node"]

        if self.centroids is not None:
            self.prepare_graph(self.centroids)
            self.set_blocked_centroid_flows(self.block_centroid_flows)
        self.__id__ = uuid.uuid4().hex

    def __build_column_names(self, all_titles: List[str]) -> Tuple[list, list]:
        fields = [x for x in self.required_default_fields]
        types = [x for x in self.__required_default_types]
        for column in all_titles:
            if column not in self.required_default_fields and column[0:-3] not in self.required_default_fields:
                if column[-3:] == "_ab":
                    if column[:-3] + "_ba" in all_titles:
                        fields.append(column[:-3])
                        types.append(self.network[column].dtype)
                    else:
                        raise ValueError("Field {} exists for ab direction but does not exist for ba".format(column))
                elif column[-3:] == "_ba":
                    if column[:-3] + "_ab" not in all_titles:
                        raise ValueError("Field {} exists for ba direction but does not exist for ab".format(column))
                else:
                    fields.append(column)
                    types.append(self.network[column].dtype)
        return fields, types

    def __build_dtype(self, all_titles) -> list:
        dtype = [
            ("link_id", self.__integer_type),
            ("a_node", self.__integer_type),
            ("b_node", self.__integer_type),
            ("direction", np.int8),
            ("id", self.__integer_type),
        ]
        for i in all_titles:
            if i not in self.required_default_fields and i[0:-3] not in self.required_default_fields:
                if i[-3:] == "_ab":
                    if i[:-3] + "_ba" in all_titles:
                        dtype.append((i[:-3], self.network[i].dtype))
                    else:
                        raise ValueError("Field {} exists for ab direction but does not exist for ba".format(i))
                elif i[-3:] == "_ba":
                    if i[:-3] + "_ab" not in all_titles:
                        raise ValueError("Field {} exists for ba direction but does not exist for ab".format(i))
                else:
                    dtype.append((i, self.network[i].dtype))
        return dtype

    def set_graph(self, cost_field) -> None:
        """
        Sets the field to be used for path computation

        :Arguments:
            **cost_field** (:obj:`str`): Field name. Must be numeric
        """
        if cost_field in self.graph.columns:
            self.cost_field = cost_field
            self.compact_cost = np.zeros(self.compact_graph.id.max() + 2, self.__float_type)
            df = self.__graph_groupby.sum()[[cost_field]].reset_index()
            self.compact_cost[df.index.values] = df[cost_field].values
            if self.graph[cost_field].dtype == self.__float_type:
                self.cost = np.array(self.graph[cost_field].values, copy=True)
            else:
                self.cost = np.array(self.graph[cost_field].values, dtype=self.__float_type)
                self.logger.warning("Cost field with wrong type. Converting to float64")
        else:
            raise ValueError("cost_field not available in the graph:" + str(self.graph.columns))

        self.__build_derived_properties()

    def set_skimming(self, skim_fields: list) -> None:
        """
        Sets the list of skims to be computed

        :Arguments:
            **skim_fields** (:obj:`list`): Fields must be numeric
        """
        if not skim_fields:
            self.skim_fields = []
            self.skims = np.array([])

        if isinstance(skim_fields, str):
            skim_fields = [skim_fields]
        elif not isinstance(skim_fields, list):
            raise ValueError("You need to provide a list of skims or the same of a single field")

        # Check if list of fields make sense
        k = [x for x in skim_fields if x not in self.graph.columns]
        if k:
            raise ValueError("At least one of the skim fields does not exist in the graph: {}".format(",".join(k)))

        self.compact_skims = np.zeros((self.compact_num_links + 1, len(skim_fields) + 1), self.__float_type)
        df = self.__graph_groupby.sum()[skim_fields].reset_index()
        for i, skm in enumerate(skim_fields):
            self.compact_skims[df.index.values, i] = df[skm].values.astype(self.__float_type)

        self.skims = np.zeros((self.num_links, len(skim_fields) + 1), self.__float_type)
        t = [x for x in skim_fields if self.graph[x].dtype != self.__float_type]
        if t:
            Warning("Some skim field with wrong type. Converting to float64")
            for i, j in enumerate(skim_fields):
                self.skims[:, i] = self.graph[j].astype(self.__float_type).values[:]
        else:
            for i, j in enumerate(skim_fields):
                self.skims[:, i] = self.graph[j].values[:]
        self.skim_fields = skim_fields

    def set_blocked_centroid_flows(self, block_centroid_flows) -> None:
        """
        Chooses whether we want to block paths to go through centroids or not.

        Default value is True

        :Arguments:
            **block_centroid_flows** (:obj:`bool`): Blocking or not
        """
        if not isinstance(block_centroid_flows, bool):
            raise TypeError("Blocking flows through centroids needs to be boolean")
        if self.num_zones == 0:
            self.logger.warning("No centroids in the model. Nothing to block")
            return
        self.block_centroid_flows = block_centroid_flows

    # Procedure to pickle graph and save to disk
    def save_to_disk(self, filename: str) -> None:
        """
        Saves graph to disk

        :Arguments:
            **filename** (:obj:`str`): Path to file. Usual file extension is *aeg*
        """
        mygraph = {}
        mygraph["description"] = self.description
        mygraph["num_links"] = self.num_links
        mygraph["num_nodes"] = self.num_nodes
        mygraph["network"] = self.network
        mygraph["graph"] = self.graph
        mygraph["all_nodes"] = self.all_nodes
        mygraph["nodes_to_indices"] = self.nodes_to_indices
        mygraph["num_nodes"] = self.num_nodes
        mygraph["fs"] = self.fs
        mygraph["cost"] = self.cost
        mygraph["cost_field"] = self.cost_field
        mygraph["skims"] = self.skims
        mygraph["skim_fields"] = self.skim_fields
        mygraph["block_centroid_flows"] = self.block_centroid_flows
        mygraph["centroids"] = self.centroids
        mygraph["graph_id"] = self.__id__
        mygraph["mode"] = self.mode

        with open(filename, "wb") as f:
            pickle.dump(mygraph, f)

    def load_from_disk(self, filename: str) -> None:
        """
        Loads graph from disk

        :Arguments:
            **filename** (:obj:`str`): Path to file
        """
        with open(filename, "rb") as f:
            mygraph = pickle.load(f)
            self.description = mygraph["description"]
            self.num_links = mygraph["num_links"]
            self.num_nodes = mygraph["num_nodes"]
            self.network = mygraph["network"]
            self.graph = mygraph["graph"]
            self.all_nodes = mygraph["all_nodes"]
            self.nodes_to_indices = mygraph["nodes_to_indices"]
            self.num_nodes = mygraph["num_nodes"]
            self.fs = mygraph["fs"]
            self.cost = mygraph["cost"]
            self.cost_field = mygraph["cost_field"]
            self.skims = mygraph["skims"]
            self.skim_fields = mygraph["skim_fields"]
            self.block_centroid_flows = mygraph["block_centroid_flows"]
            self.centroids = mygraph["centroids"]
            self.__id__ = mygraph["graph_id"]
            self.mode = mygraph["mode"]
        self.__build_derived_properties()

    def __build_derived_properties(self):
        if self.centroids is None:
            return
        self.num_zones = self.centroids.shape[0] if self.centroids.shape else 0

    def available_skims(self) -> List[str]:
        """
        Returns graph fields that are available to be set as skims

        :Returns:
            **list** (:obj:`str`): Field names
        """
        return [x for x in self.graph.columns if x not in ["link_id", "a_node", "b_node", "direction", "id"]]

    # We check if all minimum fields are there
    def __network_error_checking__(self):
        # Checking field names
        has_fields = self.network.columns
        must_fields = ["link_id", "a_node", "b_node", "direction"]
        for field in must_fields:
            if field not in has_fields:
                raise ValueError(f"could not find field {field} in the network array")

        # Uniqueness of the id
        link_ids = self.network["link_id"].astype(int)
        if link_ids.shape[0] != np.unique(link_ids).shape[0]:
            raise ValueError('"link_id" field not unique')

            # Direction values
        if np.max(self.network["direction"]) > 1 or np.min(self.network["direction"]) < -1:
            raise ValueError('"direction" field not limited to (-1,0,1) values')

        if "id" not in self.network.columns:
            self.network = self.network.assign(id=np.nan)

    def __determine_types__(self, new_type, current_type):
        if new_type.isdigit():
            new_type = int(new_type)
        else:
            try:
                new_type = float(new_type)
            except ValueError as verr:
                self.logger.warning("Could not convert {} - {}".format(new_type, verr.__str__()))
        nt = type(new_type)
        def_type = None
        if nt == int:
            def_type = int
            if current_type == float:
                def_type == float
            elif current_type == str:
                def_type = str
        elif nt == float:
            def_type = float
            if current_type == str:
                def_type = str
        elif nt == str:
            def_type = str
        else:
            raise ValueError("WRONG TYPE OR NULL VALUE")
        return def_type

    def save_compressed_correspondence(self, path, mode_name, mode_id):
        """Save graph and nodes_to_indices to disk"""
        graph_path = join(path, f"correspondence_c{mode_name}_{mode_id}.feather")
        self.graph.to_feather(graph_path)
        node_path = join(path, f"nodes_to_indices_c{mode_name}_{mode_id}.feather")
        pd.DataFrame(self.nodes_to_indices, columns=["node_index"]).to_feather(node_path)



from sqlite3 import IntegrityError
from unittest import TestCase
import string
import random
import os
from shutil import copytree, rmtree
import tempfile
import uuid
from aequilibrae.project import Project
from ...data import no_triggers_project


class TestLinkType(TestCase):
    def setUp(self) -> None:
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        copytree(no_triggers_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.curr = self.proj.conn.cursor()

        letters = [random.choice(string.ascii_letters + "_") for x in range(20)]
        self.random_string = "".join(letters)

    def tearDown(self) -> None:
        self.proj.close()

    def test_changing_link_type_id(self):
        ltypes = self.proj.network.link_types

        lt = random.choice([x for x in ltypes.all_types().values()])

        with self.assertRaises(ValueError):
            lt.link_type_id = "test my description"

        with self.assertRaises(ValueError):
            lt.link_type_id = "K"

    def test_empty(self):
        ltypes = self.proj.network.link_types

        newt = ltypes.new("Z")
        # a.link_type = 'just a_test'
        with self.assertRaises(IntegrityError):
            newt.save()

    def test_save(self):
        ltypes = self.proj.network.link_types

        newt = ltypes.new("Z")
        newt.link_type = self.random_string
        newt.description = self.random_string[::-1]
        newt.save()

        self.curr.execute('select description, link_type from link_types where link_type_id="Z"')

        desc, mname = self.curr.fetchone()
        self.assertEqual(desc, self.random_string[::-1], "Didn't save the mode description correctly")
        self.assertEqual(mname, self.random_string, "Didn't save the mode name correctly")



from unittest import TestCase
from aequilibrae.paths.vdf import VDF


class TestVDF(TestCase):
    def test_functions_available(self):
        v = VDF()
        self.assertEqual(
            v.functions_available(), ["bpr", "bpr2", "conical", "inrets"], "VDF class returning wrong availability"
        )
        self.assertEqual(v.apply_vdf, None, "VDF is missing term")
        self.assertEqual(v.apply_derivative, None, "VDF is missing term")

        with self.assertRaises(ValueError):
            v.function = "Cubic"

        with self.assertRaises(AttributeError):
            v.apply_vdf = isinstance



from unittest import TestCase
import string
import random
import os
from shutil import copytree, rmtree
import tempfile
import uuid
from aequilibrae.project import Project
from ...data import no_triggers_project


class TestLinkTypes(TestCase):
    def setUp(self) -> None:
        self.temp_proj_folder = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        copytree(no_triggers_project, self.temp_proj_folder)
        self.proj = Project()
        self.proj.open(self.temp_proj_folder)
        self.curr = self.proj.conn.cursor()

        letters = [random.choice(string.ascii_letters + "_") for x in range(20)]
        self.random_string = "".join(letters)

    def tearDown(self) -> None:
        self.proj.close()

    def test_add(self):
        lt = self.proj.network.link_types
        existing = list(lt.all_types().keys())

        newlt = lt.new("G")
        newlt.link_type = "unique_link_type"
        newlt.save()

        nowexisting = list(lt.all_types().keys())

        n = [x for x in nowexisting if x not in existing][0]
        self.assertEqual("G", n, "Failed to add link type")

    def test_delete(self):
        lt = self.proj.network.link_types
        existing = list(lt.all_types().keys())
        deleted = random.choice(existing)
        lt.delete(deleted)
        remaining = list(lt.all_types().keys())

        difference = [x for x in existing if x not in remaining]

        self.assertEqual(deleted, difference[0], "Failed to delete link type")

    def test_get_and_get_by_name(self):
        lt = self.proj.network.link_types
        ltget = lt.get("y")
        ltgetbn = lt.get_by_name("default")

        self.assertEqual(ltget.link_type_id, ltgetbn.link_type_id, "Get methods returned different things")
        self.assertEqual(ltget.link_type, ltgetbn.link_type, "Get methods returned different things")
        self.assertEqual(ltget.description, ltgetbn.description, "Get methods returned different things")
        self.assertEqual(ltget.lanes, ltgetbn.lanes, "Get methods returned different things")
        self.assertEqual(ltget.lane_capacity, ltgetbn.lane_capacity, "Get methods returned different things")
        self.assertEqual(ltget.link_type, "default", "Get methods returned different things")

    def test_all_types(self):
        lt = self.proj.network.link_types
        all_lts = set([x for x in lt.all_types().keys()])

        c = self.proj.conn.cursor()
        c.execute("select link_type_id from link_types")
        reallts = set([x[0] for x in c.fetchall()])

        diff = all_lts.symmetric_difference(reallts)
        self.assertEqual(diff, set(), "Getting all link_types failed")



from unittest import TestCase
from tempfile import gettempdir
import os
import uuid
from shutil import copytree
from aequilibrae.project import Project
from aequilibrae.project.project_creation import remove_triggers, add_triggers
from ...data import siouxfalls_project
from shapely.geometry import LineString, Point


class TestNetworkTriggers(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]
        self.proj_path = os.path.join(gettempdir(), f"aeq_{uuid.uuid4().hex}")
        copytree(siouxfalls_project, self.proj_path)
        self.siouxfalls = Project()
        self.siouxfalls.open(self.proj_path)
        remove_triggers(self.siouxfalls.conn, self.siouxfalls.logger, db_type="network")
        add_triggers(self.siouxfalls.conn, self.siouxfalls.logger, db_type="network")

    def tearDown(self) -> None:
        self.siouxfalls.close()

    def test_delete_links_delete_nodes(self):
        items = self.siouxfalls.network.count_nodes()
        self.assertEqual(24, items, "Wrong number of nodes found")
        links = self.siouxfalls.network.links
        nodes = self.siouxfalls.network.nodes

        node = nodes.get(1)
        node.is_centroid = 0
        node.save()

        # We have completely disconnected 2 nodes (1 and 2)
        for i in [1, 2, 3, 4, 5, 14]:
            link = links.get(i)
            link.delete()
        # Since node 1 is no longer a centroid, we should have only 23 nodes in the network
        items = self.siouxfalls.network.count_nodes()
        self.assertEqual(23, items, "Wrong number of nodes found")

    def test_add_regular_link(self):
        # Add a regular link to see if it fails when creating it
        # It happened at some point
        curr = self.siouxfalls.conn.cursor()
        data = [123456, "c", "default", LineString([Point(0, 0), Point(1, 1)]).wkb]

        sql = "insert into links (link_id, modes, link_type, geometry) Values(?,?,?,GeomFromWKB(?, 4326));"
        curr.execute(sql, data)
        self.siouxfalls.conn.commit()

    def test_add_regular_node_change_centroid_id(self):
        # Add a regular link to see if it fails when creating it
        # It happened at some point
        curr = self.siouxfalls.conn.cursor()
        network = self.siouxfalls.network
        nodes = network.count_nodes()

        data = [987654, 1, Point(0, 0).wkb]

        sql = "insert into nodes (node_id, is_centroid, geometry) Values(?,?,GeomFromWKB(?, 4326));"
        curr.execute(sql, data)
        self.siouxfalls.conn.commit()
        self.assertEqual(nodes + 1, network.count_nodes(), "Failed to insert node")

        curr.execute("Update nodes set is_centroid=0 where node_id=?", data[:1])
        self.siouxfalls.conn.commit()
        self.assertEqual(nodes, network.count_nodes(), "Failed to delete node when changing centroid flag")



from aequilibrae.utils.create_example import create_example
from aequilibrae import Parameters
import pytest


class TestCreateExample:
    # We use fixture parametrization here to create the projects so that we can do proper teardown
    # of the project, even when the test fails
    # see also: https://docs.pytest.org/en/6.2.x/fixture.html#fixture-parametrize
    @pytest.fixture(params=["nauru", "sioux_falls", "coquimbo"])
    def model_project(self, tmp_path, request):
        proj = create_example(str(tmp_path / request.param), from_model=request.param)
        yield proj
        proj.close()

    def test_create_example(self, model_project):
        par = Parameters._default
        parproj = model_project.parameters
        assert par.keys() == parproj.keys(), "Wrong parameter keys for example project"



import importlib.util as iutil
from tempfile import gettempdir
from unittest import TestCase
import os
from aequilibrae.project.network.osm_downloader import OSMDownloader
from random import random

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None


class TestOSMDownloader(TestCase):
    def setUp(self) -> None:
        os.environ["PATH"] = os.path.join(gettempdir(), "temp_data") + ";" + os.environ["PATH"]

    def test_do_work(self):
        if not self.should_do_work():
            return
        o = OSMDownloader([[0.0, 0.0, 0.1, 0.1]], ["car"])
        o.doWork()
        if o.json:
            self.fail("It found links in the middle of the ocean")

    def test_do_work2(self):
        if not self.should_do_work():
            return

        # LITTLE PLACE IN THE MIDDLE OF THE Grand Canyon North Rim
        o = OSMDownloader([[-112.185, 36.59, -112.179, 36.60]], ["car"])
        o.doWork()

        if "elements" not in o.json[0]:
            return

        if len(o.json[0]["elements"]) > 1000:
            self.fail("It found too many elements in the middle of the Grand Canyon")

        if len(o.json[0]["elements"]) < 10:
            self.fail("It found too few elements in the middle of the Grand Canyon")

    def should_do_work(self):
        thresh = 1.01 if os.environ.get("GITHUB_WORKFLOW", "ERROR") == "Code coverage" else 0.02
        return random() < thresh



from datetime import datetime
import hashlib
import logging
import zipfile
from copy import deepcopy
from os.path import splitext, basename
from typing import Dict
import importlib.util as iutil

import numpy as np
import pandas as pd
import pyproj
from pyproj import Transformer
from shapely.geometry import LineString

from aequilibrae.transit.constants import AGENCY_MULTIPLIER
from aequilibrae.transit.functions.get_srid import get_srid
from aequilibrae.transit.column_order import column_order
from aequilibrae.transit.date_tools import to_seconds, create_days_between, format_date
from aequilibrae.transit.parse_csv import parse_csv
from aequilibrae.transit.transit_elements import Fare, Agency, FareRule, Service, Trip, Stop, Route
from aequilibrae.utils.worker_thread import WorkerThread

spec = iutil.find_spec("PyQt5")
pyqt = spec is not None
if pyqt:
    from PyQt5.QtCore import pyqtSignal as SignalImpl
else:

    class SignalImpl:
        def __init__(self, *args, **kwargs):
            pass

        def emit(*args, **kwargs):
            pass


class GTFSReader(WorkerThread):
    """Loader for GTFS data. Not meant to be used directly by the user"""

    signal = SignalImpl(object)

    logger = logging.getLogger("GTFS Reader")

    def __init__(self):
        WorkerThread.__init__(self, None)
        self.__capacities__ = {}
        self.__max_speeds__ = {}
        self.feed_date = ""
        self.agency = Agency()
        self.services = {}
        self.routes: Dict[int, Route] = dict()
        self.trips: Dict[int, Dict[Route]] = dict()
        self.stops: Dict[int, Stop] = dict()
        self.stop_times = {}
        self.shapes = {}
        self.trip_data = {}
        self.fare_rules = []
        self.fare_attributes = {}
        self.feed_dates = []
        self.data_arrays = dict()
        self.wgs84 = pyproj.Proj("epsg:4326")
        self.srid = get_srid()
        self.transformer = Transformer.from_crs("epsg:4326", f"epsg:{self.srid}", always_xy=False)
        self.__mt = ""

    def set_feed_path(self, file_path):
        """Sets GTFS feed source to be used

        :Arguments:
            **file_path** (:obj:`str`): Full path to the GTFS feed (e.g. 'D:/project/my_gtfs_feed.zip')
        """

        self.archive_dir = file_path
        self.zip_archive = zipfile.ZipFile(self.archive_dir)
        ret = self.zip_archive.testzip()
        if ret is not None:
            self.__fail(f"GTFS feed {file_path} is not valid")

        self.__load_feed_calendar()
        self.zip_archive.close()

        self.feed_date = splitext(basename(file_path))[0]
        self.__mt = f"Reading GTFS for {self.agency.agency}"

    def _set_capacities(self, capacities: dict):
        self.__capacities__ = capacities

    def _set_maximum_speeds(self, max_speeds: dict):
        self.__max_speeds__ = max_speeds

    def load_data(self, service_date: str):
        """Loads the data for a respective service date.

        :Arguments:
            **service_date** (:obj:`str`): service date. e.g. "2020-04-01".
        """
        ag_id = self.agency.agency
        self.logger.info(f"Loading data for {service_date} from the {ag_id} GTFS feed. This may take some time")

        self.__mt = f"Reading GTFS for {ag_id}"
        self.signal.emit(["start", "master", 6, self.__mt, self.__mt])

        self.__load_date()

        self.finished()

    def finished(self):
        self.signal.emit(["finished_static_gtfs_procedure"])

    def __load_date(self):
        self.logger.debug("Starting __load_date")
        self.zip_archive = zipfile.ZipFile(self.archive_dir)
        self.__load_routes_table()
        self.__load_stops_table()
        self.__load_stop_times()
        self.__load_trips_table()
        self.__deconflict_stop_times()
        self.__load_shapes_table()
        self.__load_fare_data()

        self.zip_archive.close()

    def __deconflict_stop_times(self) -> None:
        self.logger.info("Starting deconflict_stop_times")

        msg_txt = f"Interpolating stop times for {self.agency.agency}"
        self.signal.emit(["start", "secondary", len(self.trips), msg_txt, self.__mt])
        total_fast = 0
        for prog_counter, route in enumerate(self.trips):
            self.signal.emit(["update", "secondary", prog_counter + 1, msg_txt, self.__mt])
            max_speeds = self.__max_speeds__.get(self.routes[route].route_type, pd.DataFrame([]))
            for pattern in self.trips[route]:  # type: Trip
                for trip in self.trips[route][pattern]:
                    self.logger.debug(f"De-conflicting stops for route/trip {route}/{trip.trip}")
                    stop_times = self.stop_times[trip.trip]
                    if stop_times.shape[0] != len(trip.stops):
                        self.logger.error(
                            f"Trip {trip.trip_id} has a different number of stop_times than actual stops."
                        )

                    if not stop_times.arrival_time.is_monotonic_increasing:
                        stop_times.loc[stop_times.arrival_time == 0, "arrival_time"] = np.nan
                        stop_times.arrival_time.fillna(method="ffill", inplace=True)
                    diffs = np.diff(stop_times.arrival_time.values)

                    stop_geos = [self.stops[x].geo for x in trip.stops]
                    distances = np.array([x.distance(y) for x, y in zip(stop_geos[:-1], stop_geos[1:])])

                    times = np.copy(stop_times.arrival_time.values)
                    source_time = np.zeros_like(stop_times.arrival_time.values)

                    if times[-1] == times[-2]:
                        self.logger.info(f"De-conflicting stops for route/trip {route}/{trip.trip}")
                        self.logger.info("    Had conflicting stop times in its end")
                        times[-1] += 1
                        source_time[-1] = 1
                        diffs = np.diff(times)

                    to_override = np.argwhere(diffs == 0)[:, 0] + 1
                    if to_override.shape[0] > 0:
                        self.logger.info(f"De-conflicting stops for route/trip {route}/{trip.trip}")
                        self.logger.info("     Had consecutive stops with the same timestamp")
                        for i in to_override:
                            source_time[i] = 1
                            times[i:] += 1
                        diffs = np.diff(times)

                    if max_speeds.shape[0] > 0:
                        speeds = distances / diffs
                        df = pd.DataFrame(
                            {
                                "speed": speeds,
                                "max_speed": max_speeds.speed.max(),
                                "dist": distances,
                                "elapsed_time": diffs,
                                "add_time": np.zeros(diffs.shape[0], dtype=int),
                                "source_time": source_time[1:],
                            }
                        )

                        for _, rec in max_speeds.iterrows():
                            df.loc[
                                (df.dist >= rec.min_distance) & ((df.dist < rec.max_distance)), "max_speed"
                            ] = rec.speed

                        to_fix = df[df.max_speed < df.speed].index.values
                        if to_fix.shape[0] > 0:
                            self.logger.debug(f"     Trip {trip.trip} had {to_fix.shape[0]} segments too fast")
                            total_fast += to_fix.shape[0]
                            df.loc[to_fix[0] :, "source_time"] = 2
                            for i in to_fix:
                                df.loc[i:, "add_time"] += (
                                    df.elapsed_time[i] * (df.speed[i] / df.max_speed[i] - 1)
                                ).astype(int)

                            source_time[1:] = df.source_time[:]
                            times[1:] += df.add_time[:].astype(int)

                    assert min(times[1:] - times[:-1]) > 0
                    stop_times.arrival_time.values[:] = times[:].astype(int)
                    stop_times.departure_time.values[:] = times[:].astype(int)
                    stop_times.source_time.values[:] = source_time[:].astype(int)
                    trip.arrivals = stop_times.arrival_time.values
                    trip.departures = stop_times.departure_time.values

        if total_fast:
            self.logger.warning(f"There were a total of {total_fast} segments that were too fast and were corrected")

    def __load_fare_data(self):
        self.logger.debug("Starting __load_fare_data")
        fareatttxt = "fare_attributes.txt"
        self.fare_attributes = {}
        if fareatttxt in self.zip_archive.namelist():
            self.logger.debug('  Loading "fare_attributes" table')

            with self.zip_archive.open(fareatttxt, "r") as file:
                fareatt = parse_csv(file, column_order[fareatttxt])
            self.data_arrays[fareatttxt] = fareatt

            for line in range(fareatt.shape[0]):
                data = tuple(fareatt[line][list(column_order[fareatttxt].keys())])
                headers = ["fare_id", "price", "currency", "payment_method", "transfer", "transfer_duration"]
                f = Fare(self.agency.agency_id)
                f.populate(data, headers)
                if f.fare in self.fare_attributes:
                    self.__fail(f"Fare ID {f.fare} for {self.agency.agency} is duplicated")
                self.fare_attributes[f.fare] = f

        farerltxt = "fare_rules.txt"
        self.fare_rules = []
        if farerltxt not in self.zip_archive.namelist():
            return

        self.logger.debug('  Loading "fare_rules" table')

        with self.zip_archive.open(farerltxt, "r") as file:
            farerl = parse_csv(file, column_order[farerltxt])
        self.data_arrays[farerltxt] = farerl

        corresp = {}
        zone_id = self.agency.agency_id * AGENCY_MULTIPLIER + 1
        for line in range(farerl.shape[0]):
            data = tuple(farerl[line][list(column_order[farerltxt].keys())])
            fr = FareRule()
            fr.populate(data, ["fare", "route", "origin", "destination", "contains"])
            fr.fare_id = self.fare_attributes[fr.fare].fare_id
            if fr.route in self.routes:
                fr.route_id = self.routes[fr.route].route_id
            fr.agency_id = self.agency.agency_id
            for x in [fr.origin, fr.destination]:
                if x not in corresp:
                    corresp[x] = zone_id
                    zone_id += 1
            fr.origin_id = corresp[fr.origin]
            fr.destination_id = corresp[fr.destination] if fr.destination == "" else fr.destination_id
            self.fare_rules.append(fr) if fr.origin == "" else fr.origin_id

    def __load_shapes_table(self):
        self.logger.debug("Starting __load_shapes_table")

        self.logger.debug("    Loading route shapes")
        self.shapes.clear()
        shapestxt = "shapes.txt"
        if shapestxt not in self.zip_archive.namelist():
            return

        with self.zip_archive.open(shapestxt, "r") as file:
            shapes = parse_csv(file, column_order[shapestxt])

        all_shape_ids = np.unique(shapes["shape_id"]).tolist()
        msg_txt = f"Load shapes - {self.agency.agency}"
        self.signal.emit(["start", "secondary", len(all_shape_ids), msg_txt, self.__mt])

        self.data_arrays[shapestxt] = shapes
        lons, lats = self.transformer.transform(shapes[:]["shape_pt_lat"], shapes[:]["shape_pt_lon"])
        shapes[:]["shape_pt_lat"][:] = lats[:]
        shapes[:]["shape_pt_lon"][:] = lons[:]
        for i, shape_id in enumerate(all_shape_ids):
            self.signal.emit(["update", "secondary", i + 1, msg_txt, self.__mt])
            items = shapes[shapes["shape_id"] == shape_id]
            items = items[np.argsort(items["shape_pt_sequence"])]
            shape = LineString([x for x in zip(items["shape_pt_lon"], items["shape_pt_lat"])])
            self.shapes[shape_id] = shape

    def __load_trips_table(self):
        self.logger.debug("Starting __load_trips_table")

        trip_replacements = self.__load_frequencies()

        self.logger.debug('    Loading "trips" table')
        tripstxt = "trips.txt"
        with self.zip_archive.open(tripstxt, "r") as file:
            trips_array = parse_csv(file, column_order[tripstxt])
        self.data_arrays[tripstxt] = trips_array

        msg_txt = f"Load trips - {self.agency.agency}"
        self.signal.emit(["start", "secondary", trips_array.shape[0], msg_txt, self.__mt])
        if np.unique(trips_array["trip_id"]).shape[0] < trips_array.shape[0]:
            self.__fail("There are repeated trip IDs in trips.txt")

        stp_tm = self.data_arrays["stop_times.txt"]
        diff = np.setdiff1d(trips_array["trip_id"], stp_tm["trip_id"], assume_unique=False)
        if diff.shape[0] > 0:
            diff = ",".join([str(x) for x in diff.tolist()])
            msg = f"There are IDs in trips.txt without any stop on stop_times.txt -> {diff}"
            self.logger.error(msg)
            raise Exception(msg)

        incal = np.unique(list(self.services.keys()))
        diff = np.setdiff1d(trips_array["service_id"], incal, assume_unique=False)
        if diff.shape[0] > 0:
            diff = ",".join([str(x) for x in diff.tolist()])
            self.__fail(f"There are service IDs in trips.txt that are absent in the calendar -> {diff}")

        self.trips = {str(x): {} for x in np.unique(trips_array["route_id"])}

        for i, line in enumerate(trips_array):
            self.signal.emit(["update", "secondary", i + 1, msg_txt, self.__mt])
            trip = Trip()
            trip._populate(line, trips_array.dtype.names)
            trip.route_id = self.routes[trip.route].route_id
            trip.shape = self.shapes.get(trip.shape_id, trip.shape)
            all_trips = [trip]
            if trip.trip_id in trip_replacements:
                all_trips = [deepcopy(trip) for _ in range(len(trip_replacements[trip.trip_id]))]

            for trip in all_trips:
                stop_times = self.stop_times.get(trip.trip, [])
                if len(stop_times) < 2:
                    self.logger.warning(f"Trip {trip.trip} had less than two stops, so we skipped it.")
                    continue

                cleaner = stop_times.assign(seqkey=stop_times.stop.shift(-1) + "#####" + stop_times.stop)
                cleaner.drop_duplicates(["seqkey"], inplace=True, keep="first")
                stop_times = cleaner.drop(columns=["seqkey"])
                stop_times.loc[:, "arrival_time"] = stop_times.arrival_time.astype(int)
                stop_times.loc[:, "departure_time"] = stop_times.departure_time.astype(int)
                self.stop_times[trip.trip] = stop_times
                trip.stops = list(stop_times.stop_id.values)
                m = hashlib.md5()
                m.update(trip.route.encode())
                m.update("".join(stop_times.stop.values).encode())

                trip.pattern_hash = m.hexdigest()
                trip.arrivals = list(stop_times.arrival_time.values)
                trip.departures = list(stop_times.departure_time.values)
                trip.source_time = list(stop_times.source_time.values)
                self.logger.debug(f"{trip.trip} has {len(trip.stops)} stops")
                trip._stop_based_shape = LineString([self.stops[x].geo for x in trip.stops])
                trip.shape = self.shapes.get(trip.shape)
                trip.seated_capacity = self.routes[trip.route].seated_capacity
                trip.total_capacity = self.routes[trip.route].total_capacity
                self.trips[trip.route] = self.trips.get(trip.route, {})
                self.trips[trip.route][trip.pattern_hash] = self.trips[trip.route].get(trip.pattern_hash, [])
                self.trips[trip.route][trip.pattern_hash].append(trip)

    def __load_frequencies(self):
        self.logger.debug("Starting __load_frequencies")

        trip_replacements = {}
        freqtxt = "frequencies.txt"
        if freqtxt in self.zip_archive.namelist():
            self.logger.debug('    Loading "frequencies" table')

            with self.zip_archive.open(freqtxt, "r") as file:
                freqs = parse_csv(file, column_order[freqtxt])
            self.data_arrays[freqtxt] = freqs
            for line in range(freqs.shape[0]):
                trip, start_time, end_time, headway = list(freqs[line][list(column_order[freqtxt].keys())])
                if trip not in self.stop_times:
                    self.__fail(f"trip id {trip} in frequency table has no corresponding entry in trips")

                headway = int(headway)
                if headway <= 0:
                    self.__fail(f"Trip {trip} has non-positive headway on table frequencies.txt")

                start_seconds = to_seconds(start_time)
                end_seconds = to_seconds(end_time)

                template = self.stop_times.pop(trip)

                trip_replacements[trip] = []
                steps = int(((end_seconds - start_seconds) / headway) + 1)
                for step in range(steps):
                    shift = step * headway
                    new_trip = template.copy()
                    new_trip_str = f"{trip}-{new_trip.arrival_time.values[0]}".replace(":", "")
                    new_trip.loc[:, "arrival_time"] += shift
                    new_trip.loc[:, "departure_time"] += shift
                    self.stop_times[new_trip_str] = new_trip
                    trip_replacements[trip].append(new_trip_str)
        return trip_replacements

    def __load_stop_times(self):
        self.logger.debug("Starting __load_stop_times")

        self.stop_times.clear()
        self.logger.debug('    Loading "stop times" table')
        stoptimestxt = "stop_times.txt"
        with self.zip_archive.open(stoptimestxt, "r") as file:
            stoptimes = parse_csv(file, column_order[stoptimestxt])
        self.data_arrays[stoptimestxt] = stoptimes
        msg_txt = f"Load stop times - {self.agency.agency}"

        df = pd.DataFrame(stoptimes)
        for col in ["arrival_time", "departure_time"]:
            df2 = df[col].str.split(":", expand=True)
            df2.fillna(0, inplace=True)
            df2.columns = ["h", "m", "s"]
            df2.loc[df2.h.str.len() < 1, "h"] = 0
            df2.loc[df2.m.str.len() < 1, "m"] = 0
            df2.loc[df2.s.str.len() < 1, "s"] = 0
            df2 = df2.assign(sec=0)
            df2.loc[:, "sec"] = df2.h.astype(int) * 3600 + df2.m.astype(int) * 60 + df2.s.astype(int)
            stoptimes[col] = df2.sec.values

        key = np.core.defchararray.add(stoptimes["trip_id"].astype(str), np.array(["##"] * stoptimes.shape[0]))
        key = np.core.defchararray.add(key, stoptimes["stop_sequence"].astype(str))
        if np.unique(key).shape[0] < stoptimes.shape[0]:
            self.__fail("There are repeated stop_sequences for a single trip_id on stop_times.txt")

        df = pd.DataFrame(stoptimes)

        df.loc[:, "arrival_time"] == df.loc[:, ["arrival_time", "departure_time"]].max(axis=1)
        df.loc[:, "departure_time"] == df.loc[:, "arrival_time"]

        counter = df.shape[0]
        df = df.assign(other_stop=df.stop_id.shift(-1), other_trip=df.trip_id.shift(-1))
        df = df.loc[~((df.other_stop == df.stop_id) & (df.trip_id == df.other_trip)), :]
        counter -= df.shape[0]
        df.drop(columns=["other_stop", "other_trip"], inplace=True)
        df.columns = ["stop" if x == "stop_id" else x for x in df.columns]

        stops = [s.stop for s in self.stops.values()]
        stop_ids = [s.stop_id for s in self.stops.values()]
        stop_list = pd.DataFrame({"stop": stops, "stop_id": stop_ids})
        df = df.merge(stop_list, on="stop")
        df.sort_values(["trip_id", "stop_sequence"], inplace=True)
        df = df.assign(source_time=0)
        self.signal.emit(["start", "secondary", df.trip_id.unique().shape[0], msg_txt, self.__mt])
        for trip_id, data in [[trip_id, x] for trip_id, x in df.groupby(df["trip_id"])]:
            data.loc[:, "stop_sequence"] = np.arange(data.shape[0])
            self.stop_times[trip_id] = data
            counter += data.shape[0]
            self.signal.emit(["update", "secondary", counter, msg_txt, self.__mt])

    def __load_stops_table(self):
        self.logger.debug("Starting __load_stops_table")

        self.logger.debug('    Loading "stops" table')
        self.stops = {}
        stopstxt = "stops.txt"
        with self.zip_archive.open(stopstxt, "r") as file:
            stops = parse_csv(file, column_order[stopstxt])
        self.data_arrays[stopstxt] = stops

        if np.unique(stops["stop_id"]).shape[0] < stops.shape[0]:
            self.__fail("There are repeated Stop IDs in stops.txt")

        lats, lons = self.transformer.transform(stops[:]["stop_lat"], stops[:]["stop_lon"])
        stops[:]["stop_lat"][:] = lats[:]
        stops[:]["stop_lon"][:] = lons[:]

        msg_txt = f"Load stops - {self.agency.agency}"
        self.signal.emit(["start", "secondary", stops.shape[0], msg_txt, self.__mt])
        for i, line in enumerate(stops):
            self.signal.emit(["update", "secondary", i + 1, msg_txt, self.__mt])
            s = Stop(self.agency.agency_id, line, stops.dtype.names)
            s.agency = self.agency.agency
            s.srid = self.srid
            s.get_node_id()
            self.stops[s.stop_id] = s

    def __load_routes_table(self):
        self.logger.debug("Starting __load_routes_table")

        self.logger.debug('    Loading "routes" table')
        self.routes = {}
        routetxt = "routes.txt"
        with self.zip_archive.open(routetxt, "r") as file:
            routes = parse_csv(file, column_order[routetxt])
        self.data_arrays[routetxt] = routes

        if np.unique(routes["route_id"]).shape[0] < routes.shape[0]:
            self.__fail("There are repeated route IDs in routes.txt")

        msg_txt = f"Load Routes - {self.agency.agency}"
        self.signal.emit(["start", "secondary", len(routes), msg_txt, self.__mt])

        cap = self.__capacities__.get("other", [None, None, None])
        routes = pd.DataFrame(routes)
        routes = routes.assign(seated_capacity=cap[0], total_capacity=cap[1], srid=self.srid)
        for route_type, cap in self.__capacities__.items():
            routes.loc[routes.route_type == route_type, ["seated_capacity", "total_capacity"]] = cap

        for i, line in routes.iterrows():
            self.signal.emit(["update", "secondary", i + 1, msg_txt, self.__mt])
            r = Route(self.agency.agency_id)
            r.populate(line.values, routes.columns)
            self.routes[r.route] = r

    def __load_feed_calendar(self):
        self.logger.debug("Starting __load_feed_calendar")

        self.logger.debug('    Loading "calendar" table')
        self.services.clear()

        caltxt = "calendar.txt"
        with self.zip_archive.open(caltxt, "r") as file:
            calendar = parse_csv(file, column_order[caltxt])
        calendar["start_date"] = [datetime.fromisoformat(format_date(i)) for i in calendar["start_date"]]
        calendar["end_date"] = [datetime.fromisoformat(format_date(i)) for i in calendar["end_date"]]
        self.data_arrays[caltxt] = calendar
        if np.unique(calendar["service_id"]).shape[0] < calendar.shape[0]:
            self.__fail("There are repeated service IDs in calendar.txt")

        min_date = min(calendar["start_date"].tolist())
        max_date = max(calendar["end_date"].tolist())
        self.feed_dates = create_days_between(min_date, max_date)

        for line in calendar:
            service = Service()
            service._populate(line, calendar.dtype.names)
            self.services[service.service_id] = service

        self.logger.debug('    Loading "calendar dates" table')
        caldatetxt = "calendar_dates.txt"

        if caldatetxt not in self.zip_archive.namelist():
            self.logger.warning(f"{caldatetxt} not available in this feed")
            return

        with self.zip_archive.open(caldatetxt, "r") as file:
            caldates = parse_csv(file, column_order[caldatetxt])

        if caldates.shape[0] == 0:
            return

        ordercal = list(column_order[caldatetxt].keys())
        exception_inconsistencies = 0
        for line in range(caldates.shape[0]):
            service_id, sd, exception_type = list(caldates[line][ordercal])

            sd = format_date(sd)

            if service_id not in self.services:
                s = Service()
                s.service_id = service_id
                self.services[service_id] = s
                msg = "           Service ({}) exists on calendar_dates.txt but not on calendar.txt"
                self.logger.debug(msg.format(service.service_id))
                exception_inconsistencies += 1

            service = self.services[service_id]

            if exception_type == 1:
                if sd not in service.dates:
                    service.dates.append(sd)
                else:
                    exception_inconsistencies += 1
                    msg = "ignoring service ({}) addition on a day when the service is already active"
                    self.logger.debug(msg.format(service.service_id))
            elif exception_type == 2:
                if sd in service.dates:
                    _ = service.dates.remove(sd)
                else:
                    exception_inconsistencies += 1
                    msg = "ignoring service ({}) removal on a day from which the service was absent"
                    self.logger.debug(msg.format(service.service_id))
            else:
                self.__fail(f"illegal service exception type. {service.service_id}")
        if exception_inconsistencies:
            self.logger.info("    Minor inconsistencies found between calendar.txt and calendar_dates.txt")

    def __fail(self, msg: str) -> None:
        self.logger.error(msg)
        raise Exception(msg)
